"""
╔══════════════════════════════════════════════════════════════════════╗
║  library.py – Material-property database                             ║
║                                                                      ║
║  This file is released under CC0 1.0 – Public-Domain Dedication.     ║
║  You may use, copy, modify, and distribute it without restriction.   ║
║                                                                      ║
║  All OTHER files in this repository are proprietary. See LICENSE     ║
║  at the project root for full terms, conditions, and disclaimers.    ║
║                                                                      ║
║  DISCLAIMER: Data herein are provided “as is”, with no warranties.   ║
║  Verify critical values independently before use.                    ║
╚══════════════════════════════════════════════════════════════════════╝
"""
from ._emerge.material import Material, AIR, COPPER, PEC
from ._emerge.const import C0, Z0, PI, EPS0, MU0

EISO: float = (Z0/(2*PI))**0.5
EOMNI = (3*Z0/(4*PI))**0.5

##MATERIALS 
VACUUM = Material(color="#2d8cd5", opacity=0.05)

############################################################
#                         METALS                         #
############################################################

GREY = "#bfbfbf"
MET_ALUMINUM         = Material(cond=3.77e7,  color=GREY,     opacity=0.5, _metal=True, name="Aluminum")
MET_CARBON           = Material(cond=3.33e4,  color=GREY,     opacity=0.5, _metal=True, name="Carbon")
MET_CHROMIUM         = Material(cond=5.56e6,  color=GREY,     opacity=0.5, _metal=True, name="Chromium")
MET_COPPER           = Material(cond=5.8e7,   color="#62290c", opacity=1.0, _metal=True, name="Copper")
MET_GOLD             = Material(cond=4.10e7,  color="#d4af37", opacity=0.5, _metal=True, name="Gold")
MET_INDIUM           = Material(cond=6.44e6,  color=GREY,     opacity=0.5, _metal=True, name="Indium")
MET_IRIDIUM          = Material(cond=2.13e7,  color=GREY,     opacity=0.5, _metal=True, name="Iridium")
MET_IRON             = Material(cond=1.04e7,  color="#aaaaaa", opacity=0.5, _metal=True, name="Iron")
MET_LEAD             = Material(cond=4.84e6,  color=GREY,     opacity=0.5, _metal=True, name="Lead")
MET_MAGNESIUM        = Material(cond=2.38e7,  color=GREY,     opacity=0.5, _metal=True, name="Magnesium")
MET_NICKEL           = Material(cond=1.14e7,  color=GREY,     opacity=0.5, _metal=True, name="Nickel")
MET_NICHROME         = Material(cond=9.09e5,  color=GREY,     opacity=0.5, _metal=True, name="Nichrome")
MET_PALLADIUM        = Material(cond=9.42e6,  color=GREY,     opacity=0.5, _metal=True, name="Palladium")
MET_PLATINUM         = Material(cond=9.42e6,  color=GREY,     opacity=0.5, _metal=True, name="Platinum")
MET_RHODIUM          = Material(cond=2.22e7,  color=GREY,     opacity=0.5, _metal=True, name="Rhodium")
MET_SILVER           = Material(cond=6.29e7,  color=GREY,     opacity=0.5, _metal=True, name="Silver")
MET_TANTALUM         = Material(cond=6.44e6,  color=GREY,     opacity=0.5, _metal=True, name="Tantalum")
MET_TANTALUM_NITRIDE = Material(cond=3.97e5,  color=GREY,     opacity=0.5, _metal=True, name="Tantalum Nitride")
MET_TIN              = Material(cond=8.66e6,  color=GREY,     opacity=0.5, _metal=True, name="Tin")
MET_TITANIUM         = Material(cond=1.82e6,  color=GREY,     opacity=0.5, _metal=True, name="Titanium")
MET_TUNGSTEN         = Material(cond=1.79e7,  color=GREY,     opacity=0.5, _metal=True, name="Tungsten")
MET_ZINC             = Material(cond=1.76e7,  color=GREY,     opacity=0.5, _metal=True, name="Zinc")
MET_ZIRCONIUM        = Material(cond=2.44e7,  color=GREY,     opacity=0.5, _metal=True, name="Zirconium")



############################################################
#                       SEMICONDUCTORS                     #
############################################################

SEMI_SILICON       = Material(er=11.7,  tand=0.005,   color="#b4b4b4", opacity=0.5, _metal=True, name="Silicon")              # Crystalline Si
SEMI_SILICON_N     = Material(er=7.5,   tand=0.0003,  color="#a0a0a0", opacity=0.5, _metal=True, name="Silicon Nitride")      # Si₃N₄
SEMI_SILICON_OXIDE = Material(er=3.9,   tand=0.0001,  color="#e0e0e0", opacity=0.5, _metal=True, name="Silicon Dioxide")      # SiO₂
SEMI_GERMANIUM     = Material(er=16.0,  tand=0.001,   color="#787878", opacity=0.5, _metal=True, name="Germanium")
SEMI_GAAS          = Material(er=13.1,  tand=0.0016,  color="#aa8888", opacity=0.5, _metal=True, name="Gallium Arsenide")
SEMI_GA_N          = Material(er=8.9,   tand=0.002,   color="#8888cc", opacity=0.5, _metal=True, name="Gallium Nitride")
SEMI_INP           = Material(er=12.5,  tand=0.0015,  color="#cc99aa", opacity=0.5, _metal=True, name="Indium Phosphide")
SEMI_ALN           = Material(er=8.6,   tand=0.0003,  color="#ccccee", opacity=0.5, _metal=True, name="Aluminum Nitride")
SEMI_AL2O3         = Material(er=9.8,   tand=0.0002,  color="#eaeaea", opacity=0.5, _metal=True, name="Alumina")              # Al₂O₃
SEMI_SAPPHIRE      = Material(er=9.4,   tand=0.0001,  color="#ddddff", opacity=0.5, _metal=True, name="Sapphire")
SEMI_DIAMOND       = Material(er=5.5,   tand=0.00005, color="#cceeff", opacity=0.5, _metal=True, name="Diamond")              # Synthetic CVD diamond
SEMI_HBN           = Material(er=4.0,   tand=0.0001,  color="#eeeeff", opacity=0.5, _metal=True, name="Hexagonal Boron Nitride")
SEMI_SIOXNY        = Material(er=5.0,   tand=0.002,   color="#ddddee", opacity=0.5, _metal=True, name="Silicon Oxynitride")   # SiOxNy


############################################################
#                          LIQUIDS                         #
############################################################

LIQ_WATER     = Material(er=80.1,    cond=0.0,  color="#0080ff", opacity=0.3, _metal=True, name='Water')
LIQ_FERRITE   = Material(er=12.0,    ur=2000, tand=0.02, color="#994d4d", opacity=0.3, name='Ferrite')

############################################################
#                        DIELECTRICS                       #
############################################################

#  TRADEMARKS: All product names (e.g. “DUROID”) are the property of their
#  respective owners.  Use of them here does not imply any affiliation with or
#  endorsement by those owners.

DIEL_PTFE           = Material(er=2.1, tand=0.0002, color="#21912b", opacity=0.3, name='Teflon')
DIEL_POLYIMIDE      = Material(er=3.4, tand=0.02, color="#b8b8b8", name='Polyimide')
DIEL_CERAMIC        = Material(er=6.0, tand=0.001, color="#efead1", name='Ceramic')
DIEL_AD10           = Material(er=10.2, tand=0.0078, color="#21912b", opacity=0.3, name='AD10')
DIEL_AD1000         = Material(er=10.2, tand=0.0023, color="#21912b", opacity=0.3, name='AD1000')
DIEL_AD250          = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3, name='AD250')
DIEL_AD250_PIM      = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3, name='AD250 PIM')
DIEL_AD250A         = Material(er=2.50, tand=0.0015, color="#21912b", opacity=0.3, name='AD250A')
DIEL_AD250C         = Material(er=2.50, tand=0.0014, color="#21912b", opacity=0.3, name='AD250C')
DIEL_AD255          = Material(er=2.55, tand=0.0018, color="#21912b", opacity=0.3, name='AD255')
DIEL_AD255A         = Material(er=2.55, tand=0.0015, color="#21912b", opacity=0.3, name='AD255A')
DIEL_AD255C         = Material(er=2.55, tand=0.0014, color="#21912b", opacity=0.3, name='AD255C')
DIEL_AD260A         = Material(er=2.60, tand=0.0017, color="#21912b", opacity=0.3, name='AD260A')
DIEL_AD270          = Material(er=2.7, tand=0.0023, color="#21912b", opacity=0.3, name='AD270')
DIEL_AD300          = Material(er=3, tand=0.003, color="#21912b", opacity=0.3, name='AD300')
DIEL_AD300_PIM      = Material(er=3, tand=0.003, color="#21912b", opacity=0.3, name='AD300 PIM')
DIEL_AD300A         = Material(er=3.00, tand=0.002, color="#21912b", opacity=0.3, name='AD300A')
DIEL_AD300C         = Material(er=2.97, tand=0.002, color="#21912b", opacity=0.3, name='AD300C')
DIEL_AD320          = Material(er=3.2, tand=0.0038, color="#21912b", opacity=0.3, name='AD320')
DIEL_AD320_PIM      = Material(er=3.2, tand=0.003, color="#21912b", opacity=0.3, name='AD320 PIM')
DIEL_AD320A         = Material(er=3.20, tand=0.0032, color="#21912b", opacity=0.3, name='AD320A')
DIEL_AD350          = Material(er=3.5, tand=0.003, color="#21912b", opacity=0.3, name='AD350')
DIEL_AD350_PIM      = Material(er=3.5, tand=0.003, color="#21912b", opacity=0.3, name='AD350 PIM')
DIEL_AD350A         = Material(er=3.50, tand=0.003, color="#21912b", opacity=0.3, name='AD350A')
DIEL_AD410          = Material(er=4.1, tand=0.003, color="#21912b", opacity=0.3, name='AD410')
DIEL_AD430          = Material(er=4.3, tand=0.003, color="#21912b", opacity=0.3, name='AD430')
DIEL_AD450          = Material(er=4.5, tand=0.0035, color="#21912b", opacity=0.3, name='AD450')
DIEL_AD450A         = Material(er=4.5, tand=0.0035, color="#21912b", opacity=0.3, name='AD450 A')
DIEL_AD5            = Material(er=5.1, tand=0.003, color="#21912b", opacity=0.3, name='AD5')
DIEL_AD600          = Material(er=5.90, tand=0.003, color="#21912b", opacity=0.3, name='AD600')
DIEL_AR1000         = Material(er=9.8, tand=0.003, color="#21912b", opacity=0.3, name='AR1000')
DIEL_CER_10         = Material(er=10.00, tand=0.0035, color="#21912b", opacity=0.3, name='CER 10')
DIEL_CLTE           = Material(er=2.96, tand=0.0023, color="#21912b", opacity=0.3, name='CLTE')
DIEL_CLTE_AT        = Material(er=3.00, tand=0.0013, color="#21912b", opacity=0.3, name='CLTE-AT')
DIEL_CLTE_LC        = Material(er=2.94, tand=0.0025, color="#21912b", opacity=0.3, name='CLTE LC')
DIEL_CLTE_XT        = Material(er=2.94, tand=0.0012, color="#21912b", opacity=0.3, name='CLTE XT')
DIEL_COMCLAD_HF_ER2 = Material(er=2, tand=0.0025, color="#21912b", opacity=0.3, name='Comclad HF ER2')
DIEL_COMCLAD_HF_ER3 = Material(er=3, tand=0.0025, color="#21912b", opacity=0.3, name='Comclad HF ER3')
DIEL_COMCLAD_HF_ER4 = Material(er=4, tand=0.0025, color="#21912b", opacity=0.3, name='Comclad HF ER4')
DIEL_COMCLAD_HF_ER5 = Material(er=5, tand=0.0025, color="#21912b", opacity=0.3, name='Comclad HF ER5')
DIEL_COMCLAD_HF_ER6 = Material(er=6, tand=0.0025, color="#21912b", opacity=0.3, name='Comclad HF ER6')
DIEL_COPPER_CLAD_ULTEM = Material(er=3.05, tand=0.003, color="#21912b", opacity=0.3, name='Copper clad ULTEM')
DIEL_CUCLAD_217LX   = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name='Copper Clad 217LX')
DIEL_CUCLAD_233LX   = Material(er=2.33, tand=0.0013, color="#21912b", opacity=0.3, name='Cpper Clad 233LX')
DIEL_CUCLAD_250GT   = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3, name='Copper Clad 250GT')
DIEL_CUCLAD_250GX   = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3, name='Copper Clad 250GX')
DIEL_CUFLON         = Material(er=2.05, tand=0.00045, color="#21912b", opacity=0.3, name='CUFLON')
DIEL_DICLAD_522     = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3, name='DICLAD 522')
DIEL_DICLAD_527     = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3, name='DICLAD 527')
DIEL_DICLAD_870     = Material(er=2.33, tand=0.0013, color="#21912b", opacity=0.3, name='DICLAD 870')
DIEL_DICLAD_880     = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name='DICLAD 880')
DIEL_DICLAD_880_PIM = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name='DICLAD 880 PIM')
DIEL_GETEK_3p5          = Material(er=3.5, tand=0.01, color="#21912b", opacity=0.3, name='GETEK(3.5)')
DIEL_GETEK_3p8         = Material(er=3.8, tand=0.01, color="#21912b", opacity=0.3, name='GETEK(3.8)')
DIEL_IS6802_80      = Material(er=2.80, tand=0.003,  color="#21912b", opacity=0.3, name="IS6802 80")
DIEL_IS6803_00      = Material(er=3.00, tand=0.003,  color="#21912b", opacity=0.3, name="IS6803 00")
DIEL_IS6803_20      = Material(er=3.20, tand=0.003,  color="#21912b", opacity=0.3, name="IS6803 20")
DIEL_IS6803_33      = Material(er=3.33, tand=0.003,  color="#21912b", opacity=0.3, name="IS6803 33")
DIEL_IS6803_38      = Material(er=3.38, tand=0.0032, color="#21912b", opacity=0.3, name="IS6803 38")
DIEL_IS6803_45      = Material(er=3.45, tand=0.0035, color="#21912b", opacity=0.3, name="IS6803 45")
DIEL_ISOCLAD_917    = Material(er=2.17, tand=0.0013, color="#21912b", opacity=0.3, name="Isoclad 917")
DIEL_ISOCLAD_933    = Material(er=2.33, tand=0.0016, color="#21912b", opacity=0.3, name="Isoclad 933")
DIEL_ISOLA_I_TERA_MT    = Material(er=3.45, tand=0.0030, color="#3c9747",            name="Isola I Tera Mt")
DIEL_ISOLA_NELCO_4000_13 = Material(er=3.77, tand=0.008, color="#3c9747",           name="Isola Nelco 4000 13")
DIEL_MAT_25N        = Material(er=3.38, tand=0.0025, color="#21912b", opacity=0.3, name="Mat 25n")
DIEL_MAT25FR        = Material(er=3.58, tand=0.0035, color="#21912b", opacity=0.3, name="Mat25fr")
DIEL_MEGTRON6R5775  = Material(er=3.61, tand=0.004,  color="#21912b", opacity=0.3, name="Megtron6r5775")
DIEL_MERCURYWAVE_9350 = Material(er=3.5, tand=0.004, color="#21912b", opacity=0.3, name="Mercurywave 9350")
DIEL_MULTICLAD_HF   = Material(er=3.7,  tand=0.0045, color="#21912b", opacity=0.3, name="Multiclad HF")
DIEL_N_8000         = Material(er=3.5,  tand=0.011,  color="#21912b", opacity=0.3, name="Nelco N8000")
DIEL_N4350_13RF     = Material(er=3.5,  tand=0.0065, color="#21912b", opacity=0.3, name="Nelco N4350 13RF")
DIEL_N4380_13RF     = Material(er=3.8,  tand=0.007,  color="#21912b", opacity=0.3, name="Nelco N4380 13RF")
DIEL_N8000Q         = Material(er=3.2,  tand=0.006,  color="#21912b", opacity=0.3, name="Nelco N8000Q")
DIEL_N9300_13RF     = Material(er=3.0,  tand=0.004,  color="#21912b", opacity=0.3, name="Nelco N9300 13RF")
DIEL_N9320_13RF     = Material(er=3.2,  tand=0.0045, color="#21912b", opacity=0.3, name="Nelco N9320 13RF")
DIEL_N9338_13RF     = Material(er=3.38, tand=0.0046, color="#21912b", opacity=0.3, name="Nelco N9338 13RF")
DIEL_N9350_13RF     = Material(er=3.48, tand=0.0055, color="#21912b", opacity=0.3, name="Nelco N9350 13RF")
DIEL_NH9294         = Material(er=2.94, tand=0.0022, color="#21912b", opacity=0.3, name="Nelco NH9294")
DIEL_NH9300         = Material(er=3.00, tand=0.0023, color="#21912b", opacity=0.3, name="Nelco NH9300")
DIEL_NH9320         = Material(er=3.20, tand=0.0024, color="#21912b", opacity=0.3, name="Nelco NH9320")
DIEL_NH9338         = Material(er=3.38, tand=0.0025, color="#21912b", opacity=0.3, name="Nelco NH9338")
DIEL_NH9348         = Material(er=3.48, tand=0.003,  color="#21912b", opacity=0.3, name="Nelco NH9348")
DIEL_NH9350         = Material(er=3.50, tand=0.003,  color="#21912b", opacity=0.3, name="Nelco NH9350")
DIEL_NH9410         = Material(er=4.10, tand=0.003,  color="#21912b", opacity=0.3, name="Nelco NH9410")
DIEL_NH9450         = Material(er=4.50, tand=0.003,  color="#21912b", opacity=0.3, name="Nelco NH9450")
DIEL_NORCLAD        = Material(er=2.55, tand=0.0011, color="#21912b", opacity=0.3, name="Norclad")
DIEL_NX9240         = Material(er=2.40, tand=0.0016, color="#21912b", opacity=0.3, name="Nelco NX9240")
DIEL_NX9245         = Material(er=2.45, tand=0.0016, color="#21912b", opacity=0.3, name="Nelco NX9245")
DIEL_NX9250         = Material(er=2.50, tand=0.0017, color="#21912b", opacity=0.3, name="Nelco NX9250")
DIEL_NX9255         = Material(er=2.55, tand=0.0018, color="#21912b", opacity=0.3, name="Nelco NX9255")
DIEL_NX9260         = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3, name="Nelco NX9260")
DIEL_NX9270         = Material(er=2.70, tand=0.002,  color="#21912b", opacity=0.3, name="Nelco NX9270")
DIEL_NX9294         = Material(er=2.94, tand=0.0022, color="#21912b", opacity=0.3, name="Nelco NX9294")
DIEL_NX9300         = Material(er=3.00, tand=0.0023, color="#21912b", opacity=0.3, name="Nelco NX9300")
DIEL_NX9320         = Material(er=3.20, tand=0.0024, color="#21912b", opacity=0.3, name="Nelco NX9320")
DIEL_NY9208         = Material(er=2.08, tand=0.0006, color="#21912b", opacity=0.3, name="Nelco NY9208")
DIEL_NY9217         = Material(er=2.17, tand=0.0008, color="#21912b", opacity=0.3, name="Nelco NY9217")
DIEL_NY9220         = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3, name="Nelco NY9220")
DIEL_NY9233         = Material(er=2.33, tand=0.0011, color="#21912b", opacity=0.3, name="Nelco NY9233")
DIEL_POLYGUIDE      = Material(er=2.320,tand=0.0005, color="#21912b", opacity=0.3, name="Polyguide")
DIEL_RF_30          = Material(er=3.00, tand=0.0019, color="#21912b", opacity=0.3, name="Rogers RF-30")
DIEL_RF_301         = Material(er=2.97, tand=0.0018, color="#21912b", opacity=0.3, name="Rogers RF-301")
DIEL_RF_35          = Material(er=3.50, tand=0.0025, color="#21912b", opacity=0.3, name="Rogers RF-35")
DIEL_RF_35A2        = Material(er=3.50, tand=0.0015, color="#21912b", opacity=0.3, name="Rogers RF-35A2")
DIEL_RF_35P         = Material(er=3.50, tand=0.0034, color="#21912b", opacity=0.3, name="Rogers RF-35P")
DIEL_RF_35TC        = Material(er=3.50, tand=0.0011, color="#21912b", opacity=0.3, name="Rogers RF-35TC")
DIEL_RF_41          = Material(er=4.10, tand=0.0038, color="#21912b", opacity=0.3, name="Rogers RF-41")
DIEL_RF_43          = Material(er=4.30, tand=0.0033, color="#21912b", opacity=0.3, name="Rogers RF-43")
DIEL_RF_45          = Material(er=4.50, tand=0.0037, color="#21912b", opacity=0.3, name="Rogers RF-45")
DIEL_RF_60A         = Material(er=6.15, tand=0.0038, color="#21912b", opacity=0.3, name="Rogers RF-60A")
DIEL_RO3003         = Material(er=3.00, tand=0.0011, color="#21912b", opacity=0.3, name="Rogers RO3003")
DIEL_RO3006         = Material(er=6.15, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers RO3006")
DIEL_RO3010         = Material(er=10.2, tand=0.0022, color="#21912b", opacity=0.3, name="Rogers RO3010")
DIEL_RO3035         = Material(er=3.50, tand=0.0017, color="#21912b", opacity=0.3, name="Rogers RO3035")
DIEL_RO3203         = Material(er=3.02, tand=0.0016, color="#21912b", opacity=0.3, name="Rogers RO3203")
DIEL_RO3206         = Material(er=6.15, tand=0.0027, color="#21912b", opacity=0.3, name="Rogers RO3206")
DIEL_RO3210         = Material(er=10.2, tand=0.0027, color="#21912b", opacity=0.3, name="Rogers RO3210")
DIEL_RO3730         = Material(er=3.00, tand=0.0016, color="#21912b", opacity=0.3, name="Rogers RO3730")
DIEL_RO4003C        = Material(er=3.38, tand=0.0029, color="#21912b", opacity=0.3, name="Rogers RO4003C")
DIEL_RO4350B        = Material(er=3.48, tand=0.0037, color="#21912b", opacity=0.3, name="Rogers RO4350B")
DIEL_RO4350B_TX     = Material(er=3.48, tand=0.0034, color="#21912b", opacity=0.3, name="Rogers RO4350B TX")
DIEL_RO4360         = Material(er=6.15, tand=0.0038, color="#21912b", opacity=0.3, name="Rogers RO4360")
DIEL_RO4533         = Material(er=3.30, tand=0.0025, color="#21912b", opacity=0.3, name="Rogers RO4533")
DIEL_RO4534         = Material(er=3.40, tand=0.0027, color="#21912b", opacity=0.3, name="Rogers RO4534")
DIEL_RO4535         = Material(er=3.50, tand=0.0037, color="#21912b", opacity=0.3, name="Rogers RO4535")
DIEL_RO4730         = Material(er=3.00, tand=0.0033, color="#21912b", opacity=0.3, name="Rogers RO4730")
DIEL_RT_Duroid_5870    = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3, name="Rogers RT/duroid 5870")
DIEL_RT_Duroid_5880    = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3, name="Rogers RT/duroid 5880")
DIEL_RT_Duroid_5880LZ  = Material(er=1.96, tand=0.0019, color="#21912b", opacity=0.3, name="Rogers RT/duroid 5880LZ")
DIEL_RT_Duroid_6002    = Material(er=2.94, tand=0.0012, color="#21912b", opacity=0.3, name="Rogers RT/duroid 6002")
DIEL_RT_Duroid_6006    = Material(er=6.15, tand=0.0027, color="#21912b", opacity=0.3, name="Rogers RT/duroid 6006")
DIEL_RT_Duroid_6010_2LM= Material(er=10.2, tand=0.0023, color="#21912b", opacity=0.3, name="Rogers RT/duroid 6010 2LM")
DIEL_RT_Duroid_6035HTC = Material(er=3.50, tand=0.0013, color="#21912b", opacity=0.3, name="Rogers RT/duroid 6035HTC")
DIEL_RT_Duroid_6202     = Material(er=2.94, tand=0.0015, color="#21912b", opacity=0.3, name="Rogers RT/duroid 6202")
DIEL_RT_Duroid_6202PR   = Material(er=2.90, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers RT/duroid 6202PR")
DIEL_SYRON_70000_002IN_Thick = Material(er=3.4,  tand=0.0045, color="#21912b", opacity=0.3, name="Syron 70000 0.002in")
DIEL_SYRON_71000_004IN_THICK = Material(er=3.39, tand=0.005,  color="#21912b", opacity=0.3, name="Syron 71000 0.004in")
DIEL_SYRON_71000INCH         = Material(er=3.61, tand=0.006,  color="#21912b", opacity=0.3, name="Syron 71000 Inch")
DIEL_TACLAMPLUS  = Material(er=2.10, tand=0.0004, color="#21912b", opacity=0.3, name="TacLam Plus")
DIEL_TC350       = Material(er=3.50, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers TC350")
DIEL_TC600       = Material(er=6.15, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers TC600")
DIEL_THETA       = Material(er=3.85, tand=0.0123, color="#21912b", opacity=0.3, name="Theta")
DIEL_TLA_6       = Material(er=2.62, tand=0.0017, color="#21912b", opacity=0.3, name="TLA-6")
DIEL_TLC_27      = Material(er=2.75, tand=0.003,  color="#21912b", opacity=0.3, name="TLC-27")
DIEL_TLC_30      = Material(er=3.00, tand=0.003,  color="#21912b", opacity=0.3, name="TLC-30")
DIEL_TLC_32      = Material(er=3.20, tand=0.003,  color="#21912b", opacity=0.3, name="TLC-32")
DIEL_TLC_338     = Material(er=3.38, tand=0.0034, color="#21912b", opacity=0.3, name="TLC-338")
DIEL_TLC_35      = Material(er=3.50, tand=0.0037, color="#21912b", opacity=0.3, name="TLC-35")
DIEL_TLE_95      = Material(er=2.95, tand=0.0028, color="#21912b", opacity=0.3, name="TLE-95")
DIEL_TLF_34      = Material(er=3.40, tand=0.002,  color="#21912b", opacity=0.3, name="TLF-34")
DIEL_TLF_35      = Material(er=3.50, tand=0.002,  color="#21912b", opacity=0.3, name="TLF-35")
DIEL_TLG_29      = Material(er=2.87, tand=0.0027, color="#21912b", opacity=0.3, name="TLG-29")
DIEL_TLG_30      = Material(er=3.00, tand=0.0038, color="#21912b", opacity=0.3, name="TLG-30")
DIEL_TLP_3       = Material(er=2.33, tand=0.0009, color="#21912b", opacity=0.3, name="TLP-3")
DIEL_TLP_5       = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3, name="TLP-5")
DIEL_TLP_5A      = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name="TLP-5A")
DIEL_TLT_0       = Material(er=2.45, tand=0.0019, color="#21912b", opacity=0.3, name="TLT-0")
DIEL_TLT_6       = Material(er=2.65, tand=0.0019, color="#21912b", opacity=0.3, name="TLT-6")
DIEL_TLT_7       = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3, name="TLT-7")
DIEL_TLT_8       = Material(er=2.55, tand=0.0019, color="#21912b", opacity=0.3, name="TLT-8")
DIEL_TLT_9       = Material(er=2.50, tand=0.0019, color="#21912b", opacity=0.3, name="TLT-9")
DIEL_TLX_0       = Material(er=2.45, tand=0.0019, color="#21912b", opacity=0.3, name="TLX-0")
DIEL_TLX_6       = Material(er=2.65, tand=0.0019, color="#21912b", opacity=0.3, name="TLX-6")
DIEL_TLX_7       = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3, name="TLX-7")
DIEL_TLX_8       = Material(er=2.55, tand=0.0019, color="#21912b", opacity=0.3, name="TLX-8")
DIEL_TLX_9       = Material(er=2.50, tand=0.0019, color="#21912b", opacity=0.3, name="TLX-9")
DIEL_TLY_3       = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3, name="TLY-3")
DIEL_TLY_3F      = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3, name="TLY-3F")
DIEL_TLY_5       = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3, name="TLY-5")
DIEL_TLY_5_L     = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3, name="TLY-5L")
DIEL_TLY_5A      = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name="TLY-5A")
DIEL_TLY_5AL     = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3, name="TLY-5AL")
DIEL_TMM_10      = Material(er=9.20, tand=0.0022, color="#21912b", opacity=0.3, name="Rogers TMM10")
DIEL_TMM_10i     = Material(er=9.80, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers TMM10i")
DIEL_TMM_3       = Material(er=3.27, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers TMM3")
DIEL_TMM_4       = Material(er=4.50, tand=0.002,  color="#21912b", opacity=0.3, name="Rogers TMM4")
DIEL_TMM_6       = Material(er=6.00, tand=0.0023, color="#21912b", opacity=0.3, name="Rogers TMM6")

DIEL_TRF_41      = Material(er=4.10, tand=0.0035, color="#21912b", opacity=0.3, name="TRF-41")
DIEL_TRF_43      = Material(er=4.30, tand=0.0035, color="#21912b", opacity=0.3, name="TRF-43")
DIEL_TRF_45      = Material(er=4.50, tand=0.0035, color="#21912b", opacity=0.3, name="TRF-45")

DIEL_TSM_26      = Material(er=2.60, tand=0.0014, color="#21912b", opacity=0.3, name="TSM-26")
DIEL_TSM_29      = Material(er=2.94, tand=0.0013, color="#21912b", opacity=0.3, name="TSM-29")
DIEL_TSM_30      = Material(er=3.00, tand=0.0013, color="#21912b", opacity=0.3, name="TSM-30")
DIEL_TSM_DS      = Material(er=2.85, tand=0.001,  color="#21912b", opacity=0.3, name="TSM-DS")
DIEL_TSM_DS3     = Material(er=3.00, tand=0.0011, color="#21912b", opacity=0.3, name="TSM-DS3")
DIEL_ULTRALAM_2000 = Material(er=2.4, tand=0.0019, color="#21912b", opacity=0.3, name="Arlon Ultralam 2000")
DIEL_ULTRALAM_3850 = Material(er=2.9, tand=0.0025, color="#21912b", opacity=0.3, name="Arlon Ultralam 3850")
DIEL_XT_Duroid_80000_002IN_Thick = Material(er=3.23, tand=0.0035, color="#21912b", opacity=0.3, name="Rogers XT/duroid 80000 0.002in")
DIEL_XT_Duroid_8100              = Material(er=3.54, tand=0.0049, color="#21912b", opacity=0.3, name="Rogers XT/duroid 8100")
DIEL_XT_Duroid_81000_004IN_Thick = Material(er=3.32, tand=0.0038, color="#21912b", opacity=0.3, name="Rogers XT/duroid 81000 0.004in")
DIEL_TEFLON = Material(er=2.1, tand=0.0003, color='#eeeeee', opacity=0.3, name="Teflon")

DIEL_IS420_approx   = Material(er=4.5, tand=0.018, color="#CCDE30", opacity=0.3, name="ISOLA420 approximate")

# Legacy FR Materials
DIEL_FR1 = Material(er=4.8,  tand=0.025, color="#3c9747", opacity=0.3, name="FR-1 (Paper Phenolic)")
DIEL_FR2 = Material(er=4.8,  tand=0.02,  color="#3c9747", opacity=0.3, name="FR-2 (Paper Phenolic)")
DIEL_FR3 = Material(er=4.5,  tand=0.02,  color="#2b7a4b", opacity=0.3, name="FR-3 (Paper Epoxy)")
DIEL_FR4 = Material(er=4.4,  tand=0.015, color="#1e8449", opacity=0.3, name="FR-4 (Glass Epoxy)")
DIEL_FR5 = Material(er=4.2,  tand=0.012, color="#156e38", opacity=0.3, name="FR-5 (High-temp Glass Epoxy)")
DIEL_FR6 = Material(er=5.2,  tand=0.030, color="#145a32", opacity=0.3, name="FR-6 (Paper Resin, Low Thermal)")

# Magnetic Materials
MU_METAL = Material(cond=1.0e6, ur=200000, color="#666680", opacity=0.3, name="Mu-metal")

############################################################
#                           FOAMS                          #
############################################################

FOAM_ROHACELL_31  = Material(er=1.05, tand=0.0005, color="#f0e1a1", opacity=0.15, name="Rohacell 31 (PMI Foam)")
FOAM_ROHACELL_51  = Material(er=1.07, tand=0.0006, color="#f0dea0", opacity=0.15, name="Rohacell 51 (PMI Foam)")
FOAM_ROHACELL_71  = Material(er=1.10, tand=0.0007, color="#e5d199", opacity=0.15, name="Rohacell 71 (PMI Foam)")

FOAM_PEI          = Material(er=1.15, tand=0.0035, color="#e0b56f", opacity=0.15, name="PEI Foam")
FOAM_PMI          = Material(er=1.10, tand=0.0008, color="#d9c690", opacity=0.15, name="PMI Foam")
FOAM_PVC          = Material(er=1.20, tand=0.0040, color="#cccccc", opacity=0.15, name="PVC Foam")
FOAM_EPS          = Material(er=1.03, tand=0.0050, color="#f7f7f7", opacity=0.15, name="EPS Foam (Expanded Polystyrene)")
FOAM_XPS          = Material(er=1.05, tand=0.0030, color="#e0e0e0", opacity=0.15, name="XPS Foam (Extruded Polystyrene)")
FOAM_PU           = Material(er=1.10, tand=0.0080, color="#d0d0d0", opacity=0.15, name="PU Foam (Polyurethane)")
FOAM_GLAS         = Material(er=3.10, tand=0.0050, color="#888888", opacity=0.15, name="Cellular Glass")

FOAM_AIREX_C70    = Material(er=1.10, tand=0.0010, color="#f7e7a3", opacity=0.15, name="Airex C70 (PET Foam)")
FOAM_AIREX_T92    = Material(er=1.10, tand=0.0020, color="#f6d08a", opacity=0.15, name="Airex T92 (PET Foam)")
FOAM_PVC_CORECELL = Material(er=1.56, tand=0.0025, color="#aaaaaa", opacity=0.15, name="Corecell PVC Foam")



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from loguru import logger
from typing import Callable, Literal
from dataclasses import dataclass
from collections import defaultdict
from ...selection import Selection, FaceSelection
from ...cs import CoordinateSystem, Axis, GCS, _parse_axis
from ...coord import Line
from ...geometry import GeoSurface, GeoObject
from ...bc import BoundaryCondition, BoundaryConditionSet, Periodic
from ...periodic import PeriodicCell, HexCell, RectCell
from ...material import Material
from ...const import Z0, C0, EPS0, MU0

############################################################
#                     UTILITY FUNCTIONS                    #
############################################################

def _inner_product(function: Callable, x: np.ndarray, y: np.ndarray, z: np.ndarray, ax: Axis) -> float:
            Exyz = function(x,y,z)
            return np.sum(Exyz[0,:]*ax.x + Exyz[1,:]*ax.y + Exyz[2,:]*ax.z)
        
        

############################################################
#                   MAIN BC MANAGER CLASS                  #
############################################################

class MWBoundaryConditionSet(BoundaryConditionSet):

    def __init__(self, periodic_cell: PeriodicCell | None):
        super().__init__()

        self.PEC: type[PEC] = self._construct_bc(PEC)
        self.PMC: type[PMC] = self._construct_bc(PMC)
        self.AbsorbingBoundary: type[AbsorbingBoundary] = self._construct_bc(AbsorbingBoundary)
        self.ModalPort: type[ModalPort] = self._construct_bc(ModalPort)
        self.LumpedPort: type[LumpedPort] = self._construct_bc(LumpedPort)
        self.LumpedElement: type[LumpedElement] = self._construct_bc(LumpedElement)
        self.SurfaceImpedance: type[SurfaceImpedance] = self._construct_bc(SurfaceImpedance)
        self.RectangularWaveguide: type[RectangularWaveguide] = self._construct_bc(RectangularWaveguide)
        self.Periodic: type[Periodic] = self._construct_bc(Periodic)
        self.FloquetPort: type[FloquetPort] = self._construct_bc(FloquetPort)

        self._cell: PeriodicCell | None = None

    def get_conductors(self) -> list[BoundaryCondition]:
        """Returns a list of all boundary conditions that ought to be considered as a "conductor" 
        for the purpose of modal analyses.

        Returns:
            list[BoundaryCondition]: All conductor like boundary conditions
        """
        bcs = self.oftype(PEC)
        for bc in self.oftype(SurfaceImpedance):
            if bc.sigma > 10.0:
                bcs.append(bc)

        return bcs
        
    def get_type(self, bctype: Literal['PEC','ModalPort','LumpedPort','PMC','LumpedElement','RectangularWaveguide','Periodic','FloquetPort','SurfaceImpedance']) -> FaceSelection:
        tags = []
        for bc in self.boundary_conditions:
            if bctype in str(bc.__class__):
                tags.extend(bc.selection.tags)
        return FaceSelection(tags)

    def floquet_port(self, poly: GeoSurface, port_number: int) -> FloquetPort:
        if self._cell is None:
            raise ValueError('Periodic cel must be defined for this simulation.')
        if isinstance(self._cell, RectCell):
            port = self.FloquetPort(poly, port_number)
            port.width = self._cell.width
            port.height = self._cell.height
        elif isinstance(self._cell, HexCell):
            port = self.FloquetPort(poly, port_number)
            port.area = 1.0
        self._cell._ports.append(port)
        return port

    # Checks
    def _is_excited(self) -> bool:
        for bc in self.boundary_conditions:
            if not isinstance(bc, RobinBC):
                continue
            if bc._include_force:
                return True
            
        return False

############################################################
#                    BOUNDARY CONDITIONS                   #
############################################################


class PEC(BoundaryCondition):
    
    def __init__(self,
                 face: FaceSelection | GeoSurface):
        """The general perfect electric conductor boundary condition.

        The physics compiler will by default always turn all exterior faces into a PEC.

        Args:
            face (FaceSelection | GeoSurface): The boundary surface
        """
        super().__init__(face)

class PMC(BoundaryCondition):
    pass

class RobinBC(BoundaryCondition):
    
    _include_stiff: bool = False
    _include_mass: bool = False
    _include_force: bool = False
    _isabc: bool = False

    def __init__(self, selection: GeoSurface | Selection):
        """A Generalization of any boundary condition of the third kind (Robin).

        This should not be created directly. A robin boundary condition is the generalized type behind
        port boundaries, radiation boundaries etc. Since all boundary conditions of the thrid kind (Robin)
        are assembled the same, this class is used during assembly.

        Args:
            selection (GeoSurface | Selection): The boundary surface.
        """
        super().__init__(selection)
        self.v_integration: bool = False
        self.vintline: Line | None = None
    
    def get_basis(self) -> np.ndarray:
        raise NotImplementedError('This method is not implemented')
    
    def get_inv_basis(self) -> np.ndarray | None :
        raise NotImplementedError('This method is not implemented')
    
    def get_beta(self, k0: float) -> float:
        raise NotImplementedError('get_beta not implemented for Port class')
    
    def get_gamma(self, k0: float) -> complex:
        raise NotImplementedError('get_gamma not implemented for Port class')
    
    def get_Uinc(self, x_local: np.ndarray, y_local: np.ndarray, k0: float) -> np.ndarray:
        raise NotImplementedError('get_Uinc not implemented for Port class')

class PortBC(RobinBC):
    Zvac: float = Z0
    def __init__(self, face: FaceSelection | GeoSurface):
        """(DO NOT USE) A generalization of the Port boundary condition.
        
        DO NOT USE THIS TO DEFINE PORTS. This class is only indeded for 
        class inheritance and type checking. 

        Args:
            face (FaceSelection | GeoSurface): The port face
        """
        super().__init__(face)
        self.port_number: int = -1
        self.cs: CoordinateSystem = GCS
        self.selected_mode: int = 0
        self.Z0: complex | float | None = None
        self.active: bool | None = False
        self.power: float = 1.0

    @property
    def voltage(self) -> complex | None:
        return None
        
    def get_basis(self) -> np.ndarray:
        return self.cs._basis
    
    def get_inv_basis(self) -> np.ndarray:
        return self.cs._basis_inv
    
    def portZ0(self, k0: float) -> complex | float | None:
        """Returns the port characteristic impedance given a phase constant

        Args:
            k0 (float): The phase constant

        Returns:
            complex: The port impedance
        """
        return self.Z0

    def modetype(self, k0: float) -> Literal['TEM','TE','TM']:
        return 'TEM'
    
    def Zmode(self, k0: float) -> float:
        if self.modetype(k0)=='TEM':
            return self.Zvac
        elif self.modetype(k0)=='TE':
            return k0*299792458/self.get_beta(k0) * MU0
        elif self.modetype(k0)=='TM':
            return self.get_beta(k0)/(k0*299792458*EPS0)
        else:
            raise ValueError(f'Port mode type should be TEM, TE or TM but instead is {self.modetype(k0)}')
    
    def _qmode(self, k0: float) -> float:
        """Computes a mode amplitude correction factor.
        The total output power of a port as a function of the field amplitude is not constant.
        For TE and TM modes the output power depends on the mode impedance. This factor corrects
        the mode output power to 1W by scaling the E-field appropriately.

        Args:
            k0 (float): The phase constant of the simulation

        Returns:
            float: The mode amplitude correction factor.
        """
        return np.sqrt(self.Zmode(k0)/Z0)
    
    @property
    def mode_number(self) -> int:
        return self.selected_mode + 1

    def get_beta(self, k0) -> float:
        ''' Return the out of plane propagation constant. βz.'''
        return k0
    
    def get_gamma(self, k0) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        return 1j*self.get_beta(k0)
    
    def port_mode_3d(self, 
                     xs: np.ndarray,
                     ys: np.ndarray,
                     k0: float,
                     which: Literal['E','H'] = 'E') -> np.ndarray:
        raise NotImplementedError('port_mode_3d not implemented for Port class')
    
    def port_mode_3d_global(self, 
                                x_global: np.ndarray,
                                y_global: np.ndarray,
                                z_global: np.ndarray,
                                k0: float,
                                which: Literal['E','H'] = 'E') -> np.ndarray:
            xl, yl, _ = self.cs.in_local_cs(x_global, y_global, z_global)
            Ex, Ey, Ez = self.port_mode_3d(xl, yl, k0)
            Exg, Eyg, Ezg = self.cs.in_global_basis(Ex, Ey, Ez)
            return np.array([Exg, Eyg, Ezg])

class AbsorbingBoundary(RobinBC):

    _include_stiff: bool = True
    _include_mass: bool = True
    _include_force: bool = False
    _isabc: bool = True
    
    def __init__(self,
                 face: FaceSelection | GeoSurface,
                 order: int = 2,
                 origin: tuple | None = None,
                 abctype: Literal['A','B','C','D','E'] = 'B'):
        """Creates an AbsorbingBoundary condition.

        Currently only a first order boundary condition is possible. Second order will be supported later.
        The absorbing boundary is effectively a port boundary condition (Robin) with an assumption on
        the out-of-plane phase constant. For now it always assumes the free-space propagation (normal).

        Args:
            face (FaceSelection | GeoSurface): The absorbing boundary face(s)
            order (int, optional): The order (only 1 is supported). Defaults to 1.
            origin (tuple, optional): The radiation origin. Defaults to None.
        """
        super().__init__(face)
        if origin is None:
            origin = (0., 0., 0.)
        self.order: int = order
        self.origin: tuple = origin
        self.cs: CoordinateSystem = GCS
        
        self.abctype: Literal['A','B','C','D','E']  = abctype
        self.o2coeffs: tuple[float, float] = {'A': (1.0, -0.5),
                                              'B': (1.00023, -0.51555),
                                              'C': (1.03084, -0.73631),
                                              'D': (1.06103, -0.84883),
                                              'E': (1.12500, -1.00000)
                                              }
        
    def get_basis(self) -> np.ndarray:
        return np.eye(3)

    def get_inv_basis(self) -> np.ndarray | None:
        return None
    
    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''
        return k0

    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        if self.order==1:
            return 1j*k0
        
        return 1j*k0*self.o2coeffs[self.abctype][0]
    
   
@dataclass
class PortMode:
    modefield: np.ndarray
    E_function: Callable
    H_function: Callable
    k0: float
    beta: float
    residual: float
    energy: float = 0
    norm_factor: float = 1
    freq: float = 0
    neff: float = 1
    TEM: bool = True
    Z0: float = 50.0
    polarity: float = 1.0
    modetype: Literal['TEM','TE','TM'] = 'TEM'

    def __post_init__(self):
        self.neff = self.beta/self.k0
        self.energy = np.mean(np.abs(self.modefield)**2)

    def __str__(self):
        return f'PortMode(k0={self.k0}, beta={self.beta}({self.neff:.3f}))'
    
    def set_power(self, power: complex) -> None:
        self.norm_factor = np.sqrt(1/np.abs(power))
        logger.info(f'Setting port mode amplitude to: {self.norm_factor:.2f} ')

class FloquetPort(PortBC):
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = True

    def __init__(self,
                 face: FaceSelection | GeoSurface,
                 port_number: int,
                 cs: CoordinateSystem |  None = None,
                 power: float = 1.0,
                 er: float = 1.0):
        super().__init__(face)
        if cs is None:
            cs = GCS
        self.port_number: int= port_number
        self.active: bool = False
        self.power: float = power
        self.type: str = 'TEM'
        self.mode: tuple[int,int] = (1,0)
        self.cs: CoordinateSystem = cs
        self.scan_theta: float = 0
        self.scan_phi: float = 0
        self.pol_s: complex = 1.0 + 0j
        self.pol_p: complex = 0j
        self.area: float = 1
        self.width: float | None = None
        self.height: float | None = None

        if self.cs is None:
            self.cs = GCS

    def portZ0(self, k0: float | None = None) -> complex | float | None:
        return Z0

    def get_amplitude(self, k0: float) -> float:
        return 1.0
    
    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''
        return k0*np.cos(self.scan_theta)
    
    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        return 1j*self.get_beta(k0)
    
    def get_Uinc(self, x_global: np.ndarray, y_global: np.ndarray, z_global: np.ndarray, k0: float) -> np.ndarray:
        return -2*1j*self.get_beta(k0)*self.port_mode_3d_global(x_global, y_global, z_global, k0)
    
    def port_mode_3d(self, 
                     x_local: np.ndarray,
                     y_local: np.ndarray,
                     k0: float,
                     which: Literal['E','H'] = 'E') -> np.ndarray:
        ''' Compute the port mode E-field in local coordinates (XY) + Z out of plane.'''

        kx = k0*np.sin(self.scan_theta)*np.cos(self.scan_phi)
        ky = k0*np.sin(self.scan_theta)*np.sin(self.scan_phi)
        kz = k0*np.cos(self.scan_theta)
        phi = np.exp(-1j*(x_local*kx + y_local*ky))

        P = self.pol_p
        S = self.pol_s

        E0 = self.get_amplitude(k0)*np.sqrt(2*Z0/(self.area))
        Ex = E0*(-S*np.sin(self.scan_phi) - P*np.cos(self.scan_theta)*np.cos(self.scan_phi))*phi
        Ey = E0*(S*np.cos(self.scan_phi) - P*np.cos(self.scan_theta)*np.sin(self.scan_phi))*phi
        Ez = E0*(-P*E0*np.sin(self.scan_theta))*phi
        Exyz = np.array([Ex, Ey, Ez])
        return Exyz

    def port_mode_3d_global(self, 
                            x_global: np.ndarray,
                            y_global: np.ndarray,
                            z_global: np.ndarray,
                            k0: float,
                            which: Literal['E','H'] = 'E') -> np.ndarray:
        '''Compute the port mode field for global xyz coordinates.'''
        if self.cs is None:
            raise ValueError('No coordinate system is defined for this FloquetPort')
        xl, yl, _ = self.cs.in_local_cs(x_global, y_global, z_global)
        Ex, Ey, Ez = self.port_mode_3d(xl, yl, k0)
        Exg, Eyg, Ezg = self.cs.in_global_basis(Ex, Ey, Ez)
        return np.array([Exg, Eyg, Ezg])
        
class ModalPort(PortBC):
    
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = True

    def __init__(self,
                 face: FaceSelection | GeoSurface,
                 port_number: int, 
                 cs: CoordinateSystem | None = None,
                 power: float = 1,
                 TEM: bool = False,
                 mixed_materials: bool = False):
        """Generes a ModalPort boundary condition for a port that requires eigenmode solutions for the mode.

        The boundary condition requires a FaceSelection (or GeoSurface related) object for the face and a port
        number. 
        If the face coordinate system is not provided a local coordinate system will be derived automatically
        by finding the plane that spans the face nodes with minimial out-of-plane error. 

        All modal ports require the execution of a .modal_analysis() by the physics class to define
        the port mode. 

        Args:
            face (FaceSelection, GeoSurface): The port mode face
            port_number (int): The port number as an integer
            cs (CoordinateSystem, optional): The local coordinate system of the port face. Defaults to None.
            power (float, optional): The radiated power. Defaults to 1.
            TEM (bool, optional): Wether the mode should be considered as a TEM mode. Defaults to False
            mixed_materials (bool, optional): Wether the port consists of multiple different dielectrics. This requires
                A recalculation of the port mode at every frequency
        """
        super().__init__(face)

        self.port_number: int= port_number
        self.active: bool = False
        self.power: float = power
        self.alignment_vectors: list[Axis] = []

        self.selected_mode: int = 0
        self.modes: dict[float, list[PortMode]] = defaultdict(list)

        self.TEM: bool = TEM
        self.mixed_materials: bool = mixed_materials
        self.initialized: bool = False
        self._first_k0: float | None = None
        self._last_k0: float | None = None
        

        if cs is None:
            logger.info('Constructing coordinate system from normal port')
            self.cs = Axis(self.selection.normal).construct_cs() # type: ignore
        else:
            raise ValueError('No Coordinate System could be derived.')
        self._er: np.ndarray | None = None
        self._ur: np.ndarray | None = None
    
    def portZ0(self, k0: float) -> complex | float | None:
        return self.get_mode(k0).Z0
    
    def modetype(self, k0: float) -> Literal['TEM','TE','TM']:
        return self.get_mode(k0).modetype
    
    def align_modes(self, *axes: tuple | np.ndarray | Axis) -> None:
        """Set a reriees of Axis objects that define a sequence of mode field
        alignments.

        The modes will be sorted to maximize the inner product: |∬ E(x,y) · ax dS|
        
        Args:
            *axes (tuple, np.ndarray, Axis): The alignment vectors.
        """ 
        self.alignment_vectors = [_parse_axis(ax) for ax in axes]
        
    @property
    def nmodes(self) -> int:
        if self._last_k0 is None:
            raise ValueError('ModalPort is not properly configured. No modes are defined.')
        return len(self.modes[self._last_k0])
    
    @property
    def voltage(self) -> complex:
        mode = self.get_mode(0)
        return np.sqrt(mode.Z0)
        
    def sort_modes(self) -> None:
        """Sorts the port modes based on total energy
        """
        
        if len(self.alignment_vectors) > 0:
            logger.trace(f'Sorting modes based on alignment vectors: {self.alignment_vectors}')
            X, Y, Z = self.selection.sample(5)
            X = X.flatten()
            Y = Y.flatten()
            Z = Z.flatten()
            for k0, modes in self.modes.items():
                logger.trace(f'Aligning modes for k0={k0:.3f} rad/m')
                new_modes = []
                for ax in self.alignment_vectors:
                    logger.trace(f'.mode vector {ax}')
                    integrals = [_inner_product(m.E_function, X, Y, Z, ax) for m in modes]
                    integral, opt_mode = sorted([pair for pair in zip(integrals, modes)], key=lambda x: abs(x[0]), reverse=True)[0]
                    opt_mode.polarity = np.sign(integral.real)
                    logger.trace(f'Optimal mode = {opt_mode} ({integral}), polarization alignment = {opt_mode.polarity}')
                    new_modes.append(opt_mode)
                    
                self.modes[k0] = new_modes
            return
        for k0, modes in self.modes.items():
            self.modes[k0] = sorted(modes, key=lambda m: m.beta, reverse=True)

    def get_mode(self, k0: float, i=None) -> PortMode:
        """Returns a given mode solution in the form of a PortMode object.

        Args:
            i (_type_, optional): The mode solution number. Defaults to None.

        Returns:
            PortMode: The requested PortMode object
        """
        if i is None:
            i = self.selected_mode
        return self.modes[min(self.modes.keys(), key=lambda k: abs(k - k0))][i]
    
    def global_field_function(self, k0: float = 0, which: Literal['E','H'] = 'E') -> Callable:
        ''' The field function used to compute the E-field. 
        This field-function is defined in global coordinates (not local coordinates).'''
        mode = self.get_mode(k0)
        if which == 'E':
            return lambda x,y,z: mode.norm_factor * self._qmode(k0) * mode.E_function(x,y,z)*mode.polarity
        else:
            return lambda x,y,z: mode.norm_factor * self._qmode(k0) * mode.H_function(x,y,z)*mode.polarity
    
    def clear_modes(self) -> None:
        """Clear all port mode data"""
        self.modes = defaultdict(list)
        self.initialized = False

    def add_mode(self, 
                 field: np.ndarray,
                 E_function: Callable,
                 H_function: Callable,
                 beta: float,
                 k0: float,
                 residual: float,
                 TEM: bool,
                 freq: float) -> PortMode | None:
        """Add a mode function to the ModalPort

        Args:
            field (np.ndarray): The field value array
            E_function (Callable): The E-field callable
            H_function (Callable): The H-field callable
            beta (float): The out-of-plane propagation constant 
            k0 (float): The free space phase constant
            residual (float): The solution residual
            TEM (bool): Whether its a TEM mode
            freq (float): The frequency of the port mode

        Returns:
            PortMode: The port mode object.
        """
        mode = PortMode(field, E_function, H_function, k0, beta, residual, TEM=TEM, freq=freq)
        if mode.energy < 1e-4:
            logger.debug(f'Ignoring mode due to a low mode energy: {mode.energy}')
            return None
        self.modes[k0].append(mode)
        self.initialized = True

        self._last_k0 = k0
        if self._first_k0 is None:
            self._first_k0 = k0
        else:
            ref_field = self.get_mode(self._first_k0, -1).modefield
            polarity = np.sign(np.sum(field*ref_field).real)
            logger.debug(f'Mode polarity = {polarity}')
            mode.polarity = polarity

        return mode

    def get_basis(self) -> np.ndarray:
        return self.cs._basis

    def get_inv_basis(self) -> np.ndarray:
        return self.cs._basis_inv
    
    def get_beta(self, k0: float) -> float:
        mode = self.get_mode(k0)
        if mode.TEM:
            beta = mode.beta/mode.k0 * k0
        else:
            freq = k0*299792458/(2*np.pi)
            beta = np.sqrt(mode.beta**2 + k0**2 * (1-((mode.freq/freq)**2)))
        return beta

    def get_gamma(self, k0: float) -> complex:
        return 1j*self.get_beta(k0)
    
    def get_Uinc(self, x_global: np.ndarray, y_global: np.ndarray, z_global: np.ndarray, k0) -> np.ndarray:
        return -2*1j*self.get_beta(k0)*self.port_mode_3d_global(x_global, y_global, z_global, k0)
    
    def port_mode_3d(self, 
                     x_local: np.ndarray,
                     y_local: np.ndarray,
                     k0: float,
                     which: Literal['E','H'] = 'E') -> np.ndarray:
        x_global, y_global, z_global = self.cs.in_global_cs(x_local, y_local, 0*x_local)

        Egxyz = self.port_mode_3d_global(x_global,y_global,z_global,k0,which=which)
        
        Ex, Ey, Ez = self.cs.in_local_basis(Egxyz[0,:], Egxyz[1,:], Egxyz[2,:])

        Exyz = np.array([Ex, Ey, Ez])
        return Exyz

    def port_mode_3d_global(self,
                            x_global: np.ndarray,
                            y_global: np.ndarray,
                            z_global: np.ndarray,
                            k0: float,
                            which: Literal['E','H'] = 'E') -> np.ndarray:
        Ex, Ey, Ez = self.global_field_function(k0, which)(x_global,y_global,z_global)
        Exyz = np.array([Ex, Ey, Ez])
        return Exyz

class RectangularWaveguide(PortBC):
    
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = True

    def __init__(self, 
                 face: FaceSelection | GeoSurface,
                 port_number: int, 
                 mode: tuple[int, int] = (1,0),
                 cs: CoordinateSystem | None = None,
                 dims: tuple[float, float] | None = None,
                 power: float = 1):
        """Creates a rectangular waveguide as a port boundary condition.
        
        Currently the Rectangular waveguide only supports TE0n modes. The mode field
        is derived analytically. The local face coordinate system and dimensions can be provided
        manually. If not provided the class will attempt to derive the local coordinate system and
        face dimensions itself. It always orients the longest edge along the local X-direction.
        The information on the derived coordiante system will be shown in the DEBUG level logs.

        Args:
            face (FaceSelection, GeoSurface): The port boundary face selection
            port_number (int): The port number
            mode: (tuple[int, int], optional): The TE mode number. Defaults to (1,0).
            cs (CoordinateSystem, optional): The local coordinate system. Defaults to None.
            dims (tuple[float, float], optional): The port face. Defaults to None.
            power (float): The port power. Default to 1.
        """
        super().__init__(face)

        self.port_number: int= port_number
        self.active: bool = False
        self.power: float = power
        self.type: str = 'TE'
        self.mode: tuple[int,int] = mode

        if dims is None:
            logger.debug(f" - Establishing RectangularWaveguide port face based on selection {self.selection}")
            cs, (width, height) = self.selection.rect_basis() # type: ignore
            self.cs = cs # type: ignore
            self.dims = (width, height)
            logger.debug(f' - Port CS: {self.cs}')
            logger.debug(f' - Detected port {self.port_number} size = {width*1000:.1f} mm x {height*1000:.1f} mm')
        else:
            self.dims = dims
            self.cs = cs
            
        if self.cs is None:
            logger.info(' - Constructing coordinate system from normal port')
            self.cs = Axis(self.selection.normal).construct_cs()
            logger.debug(f' - Port CS: {self.cs}')
    def get_basis(self) -> np.ndarray:
        return self.cs._basis
        
    def get_inv_basis(self) -> np.ndarray:
        return self.cs._basis_inv
    
    def portZ0(self, k0: float) -> complex:
        return k0*299792458 * MU0/self.get_beta(k0)

    def modetype(self, k0):
        return self.type
    
    def get_amplitude(self, k0: float) -> float:
        Zte = Z0
        amplitude= np.sqrt(self.power*4*Zte/(self.dims[0]*self.dims[1]))
        return amplitude

    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''
        width=self.dims[0]
        height=self.dims[1]
        beta = np.sqrt(k0**2 - (np.pi*self.mode[0]/width)**2 - (np.pi*self.mode[1]/height)**2)
        return beta
    
    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        return 1j*self.get_beta(k0)
    
    def get_Uinc(self, x_global: np.ndarray, y_global: np.ndarray, z_global: np.ndarray, k0: float) -> np.ndarray:
        return -2*1j*self.get_beta(k0)*self.port_mode_3d_global(x_global, y_global, z_global, k0)
    
    def port_mode_3d(self, 
                     x_local: np.ndarray,
                     y_local: np.ndarray,
                     k0: float,
                     which: Literal['E','H'] = 'E') -> np.ndarray:
        ''' Compute the port mode E-field in local coordinates (XY) + Z out of plane.'''

        width = self.dims[0]
        height = self.dims[1]
        m, n= self.mode
        Ev = self.get_amplitude(k0)*np.cos(np.pi*m*(x_local)/width)*np.cos(np.pi*n*(y_local)/height)
        Eh = self.get_amplitude(k0)*np.sin(np.pi*m*(x_local)/width)*np.sin(np.pi*n*(y_local)/height)
        Ex = Eh
        Ey = Ev
        Ez = 0*Eh
        Exyz =  self._qmode(k0) * np.array([Ex, Ey, Ez])
        return Exyz

    def port_mode_3d_global(self, 
                            x_global: np.ndarray,
                            y_global: np.ndarray,
                            z_global: np.ndarray,
                            k0: float,
                            which: Literal['E','H'] = 'E') -> np.ndarray:
        '''Compute the port mode field for global xyz coordinates.'''
        xl, yl, _ = self.cs.in_local_cs(x_global, y_global, z_global)
        Ex, Ey, Ez = self.port_mode_3d(xl, yl, k0)
        Exg, Eyg, Ezg = self.cs.in_global_basis(Ex, Ey, Ez)
        return np.array([Exg, Eyg, Ezg])

class LumpedPort(PortBC):
    
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = True

    def __init__(self, 
                 face: FaceSelection | GeoSurface,
                 port_number: int, 
                 width: float | None = None,
                 height: float | None = None,
                 direction: Axis | tuple[float, float, float] | None = None,
                 power: float = 1,
                 Z0: float = 50):
        """Generates a lumped power boundary condition.
        
        The lumped port boundary condition assumes a uniform E-field along the "direction" axis.
        The port with and height must be provided manually in meters. The height is the size
        in the "direction" axis along which the potential is imposed. The width dimension
        is orthogonal to that. For a rectangular face its the width and for a cyllindrical face
        its the circumpherance.

        Args:
            face (FaceSelection, GeoSurface): The port surface
            port_number (int): The port number
            width (float): The port width (meters).
            height (float): The port height (meters).
            direction (Axis): The port direction as an Axis object (em.Axis(..) or em.ZAX)
            power (float, optional): The port output power. Defaults to 1.
            Z0 (float, optional): The port impedance. Defaults to 50.
        """
        super().__init__(face)

        if width is None:
            if not isinstance(face, GeoObject):
                raise ValueError(f'The width, height and direction must be defined. Information cannot be extracted from {face}')
            width, height, direction = face._data('width','height','vdir')
            if width is None or height is None or direction is None:
                raise ValueError(f'The width, height and direction could not be extracted from {face}')
        
        logger.debug(f'Lumped port: width={1000*width:.1f}mm, height={1000*height:.1f}mm, direction={direction}') # type: ignore
        self.port_number: int= port_number
        self.active: bool = False

        self.power: float = power
        self.Z0: float = Z0
        
        self.width: float = width
        self.height: float = height # type: ignore
        self.Vdirection: Axis = _parse_axis(direction) # type: ignore
        self.type = 'TEM'
        
        # logger.info('Constructing coordinate system from normal port')
        # self.cs = Axis(self.selection.normal).construct_cs()  # type: ignore
        self.cs = GCS
        self.vintline: list[Line] = []
        self.v_integration = True

    @property
    def surfZ(self) -> float:
        """The surface sheet impedance for the lumped port

        Returns:
            float: The surface sheet impedance
        """
        return self.Z0*self.width/self.height
    
    @property
    def voltage(self) -> float:
        """The Port voltage required for the provided output power (time average)

        Returns:
            float: The port voltage
        """
        return np.sqrt(2*self.power*self.Z0)
    
    def get_basis(self) -> np.ndarray:
        return self.cs._basis
        
    def get_inv_basis(self) -> np.ndarray:
        return self.cs._basis_inv
    
    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''

        return k0
    
    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        return 1j*k0*Z0/self.surfZ
    
    def get_Uinc(self, x_global: np.ndarray, y_global: np.ndarray, z_global: np.ndarray, k0) -> np.ndarray:
        Emag = -1j*2*k0 * self.voltage/self.height * (Z0/self.surfZ)
        return Emag*self.port_mode_3d_global(x_global, y_global, z_global, k0)
    
    def port_mode_3d(self, 
                     x_local: np.ndarray,
                     y_local: np.ndarray,
                     k0: float,
                     which: Literal['E','H'] = 'E') -> np.ndarray:
        ''' Compute the port mode E-field in local coordinates (XY) + Z out of plane.'''
        raise RuntimeError('This function should never be called in this context.')

    def port_mode_3d_global(self, 
                            x_global: np.ndarray,
                            y_global: np.ndarray,
                            z_global: np.ndarray,
                            k0: float,
                            which: Literal['E','H'] = 'E') -> np.ndarray:
        """Computes the port-mode field in global coordinates.

        The mode field will be evaluated at x,y,z coordinates but projected onto the local 2D coordinate system.
        Additionally, the "which" parameter may be used to request the H-field. This parameter is not always supported.

        Args:
            x_global (np.ndarray): The X-coordinate
            y_global (np.ndarray): The Y-coordinate
            z_global (np.ndarray): The Z-coordinate
            k0 (float): The free space propagation constant
            which (Literal["E","H"], optional): Which field to return. Defaults to 'E'.

        Returns:
            np.ndarray: The E-field in (3,N) indexing.
        """
        ON = np.ones_like(x_global)
        Ex, Ey, Ez = self.Vdirection.np
        return np.array([Ex*ON, Ey*ON, Ez*ON])


class LumpedElement(RobinBC):
    
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = False

    def __init__(self, 
                 face: FaceSelection | GeoSurface,
                 impedance_function: Callable | None = None,
                 width: float | None = None,
                 height: float | None = None,
                 ):
        """Generates a lumped power boundary condition.
        
        The lumped port boundary condition assumes a uniform E-field along the "direction" axis.
        The port with and height must be provided manually in meters. The height is the size
        in the "direction" axis along which the potential is imposed. The width dimension
        is orthogonal to that. For a rectangular face its the width and for a cyllindrical face
        its the circumpherance.

        Args:
            face (FaceSelection, GeoSurface): The port surface
            port_number (int): The port number
            width (float): The port width (meters).
            height (float): The port height (meters).
            direction (Axis): The port direction as an Axis object (em.Axis(..) or em.ZAX)
            active (bool, optional): Whether the port is active. Defaults to False.
            power (float, optional): The port output power. Defaults to 1.
            Z0 (float, optional): The port impedance. Defaults to 50.
        """
        super().__init__(face)

        if width is None:
            if not isinstance(face, GeoObject):
                raise ValueError(f'The width, height and direction must be defined. Information cannot be extracted from {face}')
            width, height, impedance_function = face._data('width','height','func')
            if width is None or height is None or impedance_function is None:
                raise ValueError(f'The width, height and impedance function could not be extracted from {face}')
        
        logger.debug(f'Lumped port: width={1000*width:.1f}mm, height={1000*height:.1f}mm') # type: ignore

        self.Z0: Callable = impedance_function # type: ignore
        self.width: float = width # type: ignore
        self.height: float = height # type: ignore

    def surfZ(self, k0: float) -> float:
        """The surface sheet impedance for the lumped Element

        Returns:
            float: The surface sheet impedance
        """
        Z0 = self.Z0(k0*299792458/(2*np.pi))*self.width/self.height
        return Z0
    
    def get_basis(self) -> np.ndarray | None:
        return None

    def get_inv_basis(self) -> np.ndarray | None:
        return None
    
    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''

        return k0

    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        return 1j*k0*Z0/self.surfZ(k0)
    
class SurfaceImpedance(RobinBC):
    
    _include_stiff: bool = True
    _include_mass: bool = False
    _include_force: bool = False

    def __init__(self, 
                 face: FaceSelection | GeoSurface,
                 material: Material | None = None,
                 surface_conductance: float | None = None,
                 surface_roughness: float = 0,
                 thickness: float | None = None,
                 sr_model: Literal['Hammerstad-Jensen'] = 'Hammerstad-Jensen',
                 ):
        """Generates a SurfaceImpedance bounary condition.

        The surface impedance model treats a 2D surface selection as a finite conductor. It is not
        intended to be used for dielectric materials.

        The surface resistivity is computed based on the material properties: σ, ε and μ. 

        The user may also supply the surface condutivity directly. 

        Optionally, a surface roughness in meters RMS may be supplied. In the current implementation
        The Hammersstad-Jensen model is used increasing the resistivity by a factor (1 + 2/π tan⁻¹(1.4(Δ/δ)²).

        Args:
            face (FaceSelection | GeoSurface): The face to apply this condition to.
            material (Material | None, optional): The matrial to assign. Defaults to None.
            surface_conductance (float | None, optional): The specific bulk conductivity to use. Defaults to None.
            surface_roughness (float, optional): The surface roughness. Defaults to 0.
            thickness (float | None, optional): The layer thickness. Defaults to None
            sr_model (Literal["Hammerstad-Jensen", optional): The surface roughness model. Defaults to 'Hammerstad-Jensen'.
        """
        super().__init__(face)

        self._material: Material | None = material
        self._mur: float | complex = 1.0
        self._epsr: float | complex = 1.0
        self.sigma: float = 0.0
        self.thickness: float | None = thickness
        
        if isinstance(face, GeoObject) and thickness is None:
            self.thickness = face._load('thickness')
            
        if material is not None:
            self.sigma = material.cond.scalar(1e9)
            self._mur = material.ur
            self._epsr = material.er
        
        if surface_conductance is not None:
            self.sigma = surface_conductance
        
        self._sr: float = surface_roughness
        self._sr_model: str = sr_model
    
    def get_basis(self) -> np.ndarray | None:
        return None

    def get_inv_basis(self) -> np.ndarray | None:
        return None
    
    def get_beta(self, k0: float) -> float:
        ''' Return the out of plane propagation constant. βz.'''

        return k0

    def get_gamma(self, k0: float) -> complex:
        """Computes the γ-constant for matrix assembly. This constant is required for the Robin boundary condition.

        Args:
            k0 (float): The free space propagation constant.

        Returns:
            complex: The γ-constant
        """
        
        w0 = k0*C0
        f0 = w0/(2*np.pi)
        sigma = self.sigma
        mur = self._material.ur.scalar(f0)
        er = self._material.er.scalar(f0)
        eps = EPS0*er
        mu = MU0*mur
        rho = 1/sigma
        d_skin = (2*rho/(w0*mu) * ((1+(w0*eps*rho)**2)**0.5 + rho*w0*eps))**0.5
        logger.debug(f'Computed skin depth δ={d_skin*1e6:.2}μm')
        R = (1+1j)*rho/d_skin
        if self.thickness is not None:
            eps_c = eps - 1j * sigma / w0
            gamma_m = 1j * w0 * np.sqrt(mu*eps_c)
            R = R / np.tanh(gamma_m * self.thickness)
            logger.debug(f'Impedance scaler due to thickness: {1/ np.tanh(gamma_m * self.thickness) :.4f}')
        if self._sr_model=='Hammerstad-Jensen' and self._sr > 0.0:
            R = R * (1 + 2/np.pi * np.arctan(1.4*(self._sr/d_skin)**2))
        return 1j*k0*Z0/R



import sys
import os
import ast
import inspect

def called_from_main_function() -> bool:
    """This function checks if the main simulation file is ran from if __name__ == '__main__':

    Returns:
        bool: If the script is run as main function
    """
    # 1) Locate the __main__ module’s file
    main_mod = sys.modules.get("__main__")
    main_file = getattr(main_mod, "__file__", None)
    if not main_file:
        return False
    main_file = os.path.realpath(main_file)

    # 2) Parse its AST to find the `main` function and its line range
    try:
        with open(main_file, "r") as f:
            src = f.read()
    except OSError:
        return False

    try:
        tree = ast.parse(src, filename=main_file)
    except SyntaxError:
        return False

    main_ranges = []
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name == "main":
            start = node.lineno
            # end_lineno is available in Python 3.8+; otherwise compute it
            end = getattr(node, "end_lineno",
                          max(getattr(n, "lineno", start) for n in ast.walk(node)))
            main_ranges.append((start, end))

    if not main_ranges:
        return False

    # 3) Inspect the call stack for any frame inside main_file within those line ranges
    for frame_info in inspect.stack():
        fn = os.path.realpath(frame_info.filename)
        if fn != main_file:
            continue
        lineno = frame_info.lineno
        for start, end in main_ranges:
            if start <= lineno <= end:
                return True

    return False




import re
from typing import Any
from ..pcb import PCB, RouteException
from ...material import Material, PEC
from ...cs import GCS, CoordinateSystem
from loguru import logger
from math import hypot

try: 
    import ezdxf
    from ezdxf.recover import readfile as recover_readfile
    from ezdxf.path import make_path, from_hatch
except ImportError as e:
    logger.error('Cannot find the required ezdxf library. Install using: pip install ezdxf')
    raise e

INSUNITS_TO_NAME = {
    0: "unitless",
    1: "inch",
    2: "foot",
    3: "mile",
    4: "millimeter",
    5: "centimeter",
    6: "meter",
    7: "kilometer",
    8: "microinch",
    9: "mil",          # thousandth of an inch
    10: "yard",
    11: "angstrom",
    12: "nanometer",
    13: "micrometer",
    14: "decimeter",
    15: "decameter",
    16: "hectometer",
    17: "gigameter",
    18: "astronomical unit",
    19: "light year",
    20: "parsec",
}

# scale factor: drawing units -> millimeters
INSUNITS_TO_MM = {
    0: 1.0,           # unitless; treat as mm by convention
    1: 25.4,
    2: 304.8,
    3: 1609344.0,
    4: 1.0,
    5: 10.0,
    6: 1000.0,
    7: 1_000_000.0,
    8: 25.4e-6,
    9: 0.0254,
    10: 914.4,
    11: 1e-7,
    12: 1e-6,
    13: 1e-3,
    14: 100.0,
    15: 10_000.0,
    16: 100_000.0,
    17: 1e12,
    18: 1.495978707e14,     # AU in mm
    19: 9.460730472e18,     # ly in mm
    20: 3.085677581e19,     # pc in mm
}

def cluster_values(values, tol):
    """Return [(center, count), ...] clustering sorted values by tolerance."""
    if not values:
        return []
    values = sorted(values)
    clusters = [[values[0]]]
    for v in values[1:]:
        if abs(v - clusters[-1][-1]) <= tol:
            clusters[-1].append(v)
        else:
            clusters.append([v])
    out = []
    for c in clusters:
        out.append((sum(c)/len(c), len(c)))
    return out

STACKUP_TOKENS = [
    r"fr-?4", r"rogers ?\d+", r"core", r"prepreg", r"dielectric",
    r"cu(?:\W|$)|copper", r"\b1\.?6\s*mm\b", r"\b0\.?8\s*mm\b",
    r"\b[12]\s*oz\b", r"\b35\s*µ?m\b", r"\b70\s*µ?m\b",
    r"stack[- ]?up", r"layer\s*\d+", r"pcb", r"thickness",
]

def inspect_pcb_from_dxf(
    filename: str,
    flatten_tol: float = 0.25,    # curve flattening tolerance (drawing units)
    z_tol: float | None = None,   # cluster tolerance for Z (drawing units); default auto
    layer_filter: str | None = None,
):
    """
    Returns a dict with:
      {
        'units': {'code': int, 'name': str, 'to_mm': float},
        'z_levels': [{'z': float, 'count': int}],           # in drawing units
        'thickness_units': 'same as units',
        'thickness': float,                                  # in drawing units
        'thickness_mm': float,                               # convenience
        'notes': {'materials': [...], 'raw_text_hits': [...]}
      }
    """
    doc, auditor = recover_readfile(filename)
    msp = doc.modelspace()

    # Units
    code = int(doc.header.get("$INSUNITS", 0))
    unit_name = INSUNITS_TO_NAME.get(code, f"unknown({code})")
    to_mm = INSUNITS_TO_MM.get(code, 1.0)

    # Gather geometry Zs (WCS), also consider entity.dxf.elevation as fallback
    z_values = []

    def on_layer(e) -> bool:
        return layer_filter is None or e.dxf.layer == layer_filter

    for e in msp:
        if not on_layer(e):
            continue
        dxtype = e.dxftype()
        paths = []
        try:
            if dxtype in ("HATCH", "MPOLYGON"):
                paths = list(from_hatch(e))
            else:
                paths = [make_path(e)]
        except TypeError:
            # skip unsupported entity types
            pass
        except Exception:
            pass

        for p in paths:
            subs = p.sub_paths() if getattr(p, "has_sub_paths", False) and p.has_sub_paths else [p]
            for sp in subs:
                try:
                    verts = list(sp.flattening(distance=flatten_tol))
                except Exception:
                    continue
                for v in verts:
                    # v may be Vec2 or Vec3; handle both
                    z = getattr(v, "z", None)
                    if z is None:
                        # fallback to entity elevation if available
                        z = float(getattr(e.dxf, "elevation", 0.0))
                    z_values.append(float(z))

        # Entities with explicit THICKNESS (rare but possible)
        thick = getattr(e.dxf, "thickness", None)
        if thick not in (None, 0):
            # If an entity is extruded, its "top" would be at elevation+thickness along normal.
            # We can't reliably map OCS normal here; just record the magnitude as a hint.
            pass

    # Cluster Zs
    auto_tol = (1e-6 if to_mm >= 1.0 else 1e-6 / to_mm)  # ~1 nm in mm space → tiny in most units
    ztol = z_tol if z_tol is not None else auto_tol
    z_clusters = cluster_values(z_values, tol=ztol)
    z_levels = [{"z": z, "count": n} for z, n in z_clusters]

    # Thickness from geometry (only meaningful if multiple distinct Zs)
    if z_values:
        zmin, zmax = min(z_values), max(z_values)
        thickness = zmax - zmin
    else:
        zmin = zmax = thickness = 0.0

    # Parse material/thickness hints from TEXT/MTEXT
    material_hits = set()
    raw_text_hits = []
    token_re = re.compile("|".join(STACKUP_TOKENS), re.IGNORECASE)

    for e in msp.query("TEXT MTEXT"):
        try:
            text = e.dxf.text if e.dxftype() == "TEXT" else e.text
        except Exception:
            continue
        if not text:
            continue
        if token_re.search(text):
            raw_text_hits.append(text.strip())
            # simple keyword extraction
            for kw in ["FR4", "Rogers", "core", "prepreg", "copper", "stackup", "thickness", "oz", "µm", "um", "mm"]:
                if re.search(kw, text, re.IGNORECASE):
                    material_hits.add(kw.upper())

    return {
        "units": {"code": code, "name": unit_name, "to_mm": to_mm},
        "z_levels": z_levels,
        "thickness_units": unit_name,
        "thickness": thickness,
        "thickness_mm": thickness * to_mm,
        "notes": {
            "materials": sorted(material_hits),
            "raw_text_hits": raw_text_hits[:50],  # cap to keep output sane
            "comment": (
                "Single Z level detected; geometry alone cannot determine PCB thickness."
                if len(z_levels) <= 1 else
                "Thickness estimated from min/max Z across all geometry."
            )
        },
    }

def path_items_with_semantics(e, p, distance=0.25):
    """Yield dicts describing each sub-path with geometry + semantics we can infer."""
    subs = p.sub_paths() if getattr(p, "has_sub_paths", False) and p.has_sub_paths else [p]
    for sp in subs:
        verts = list(sp.flattening(distance=distance))
        if len(verts) < 2:
            continue

        # geometry in XY (keep Z if you like)
        pts = [(v.x, v.y) for v in verts]
        is_closed = pts[0] == pts[-1]

        item = {
            "layer": e.dxf.layer,
            "entity": e.dxftype(),
            "handle": e.dxf.handle,
            "is_closed": is_closed,
            "points": pts if is_closed else None,       # polygon ring for closed shapes
            "start": None,
            "end": None,
            "length": None,                             # centerline length for open paths
            "width": None,                              # LWPOLYLINE width if available
        }

        if not is_closed:
            item["start"] = pts[0]
            item["end"]   = pts[-1]
            # polyline centerline length
            item["length"] = sum(
                hypot(x2 - x1, y2 - y1) for (x1, y1), (x2, y2) in zip(pts, pts[1:])
            )

        # Preserve potential trace width info from LWPOLYLINE
        if e.dxftype() == "LWPOLYLINE":
            cw = getattr(e.dxf, "const_width", None)
            if cw not in (None, 0):
                item["width"] = float(cw)
            else:
                # per-vertex widths (rare but possible)
                try:
                    item["widths"] = [(v[0], v[1]) for v in zip(e.get_start_widths(), e.get_end_widths())]
                except Exception:
                    pass

        yield item
        
def extract_polygons_with_meta(
    filename: str,
    layer: str | None = None,
    distance: float = 0.25,
    skip_open: bool = True,
) -> dict[str, Any]:
    """
    Returns a list of dicts:
      {
        'ring': [(x, y), ...],         # closed ring
        'layer': 'LayerName',
        'z': 12.34,                    # representative height (median Z of ring)
        'entity': 'LWPOLYLINE' | ...,
        'handle': 'ABCD'
      }
    """
    doc, auditor = recover_readfile(filename)
    msp = doc.modelspace()

    def on_layer(e) -> bool:
        return layer is None or e.dxf.layer == layer

    items = []

    for e in msp:
        if not on_layer(e):
            continue

        dxtype = e.dxftype()
        paths = []
        try:
            if dxtype in ("HATCH", "MPOLYGON"):
                paths = list(from_hatch(e))
            else:
                paths = [make_path(e)]
        except TypeError:
            continue
        except Exception:
            continue

        for p in paths:
            subs = p.sub_paths() if p.has_sub_paths else [p]
            for sp in subs:
                
                verts = list(sp.flattening(distance=distance))  # Vec3 in WCS
                if len(verts) < 2:
                    continue

                # ensure not closed
                if (verts[0].x, verts[0].y) == (verts[-1].x, verts[-1].y):
                    verts = verts[:-1]

                # ring in XY (your original target), but also compute a Z for reference
                ring_xy = [(v.x, v.y) for v in verts]
                zs = sorted(v.z for v in verts)
                z = zs[len(zs)//2] if zs else float(getattr(e.dxf, "elevation", 0.0))  # median Z

                items.append({
                    "ring": ring_xy,
                    "layer": e.dxf.layer,
                    "z": float(z),
                    "entity": dxtype,
                    "handle": e.dxf.handle,
                })

    return items


def import_dxf(filename: str, 
               material: Material, 
               thickness: float | None = None,
               unit: float | None = None,
               cs: CoordinateSystem | None = GCS,
               trace_material: Material = PEC) -> PCB:
    
    polies = extract_polygons_with_meta(filename)
    prop = inspect_pcb_from_dxf(filename)
    
    if prop['units']['name'] == 'unitless':
        if unit is None:
            raise RouteException(f'Cannot generate PCB because the unit is not found in the DXF file or provided in the import_dxf function.')
        pcb_unit = unit
    else:
        pcb_unit = 0.001 * prop['units']['to_mm']
    
    if prop['thickness'] == 0.0:
        if thickness is None:
            raise RouteException(f'Cannot generate PCB because no thickness is found int he DXF file and none is provided in the import_dxf function.')
        pcb_thickness = thickness
    else:
        pcb_thickness = 0.001 * prop['thickness_mm'] / pcb_unit
    
    if cs is None:
        cs = GCS
    
    zs = sorted(list(set([pol['z'] for pol in polies])))
    pcb = PCB(pcb_thickness, pcb_unit, cs, material=material, trace_material=trace_material)
    
    for poly in polies:
        xs, ys = zip(*poly['ring'])
        z = poly['z']
        zs.append(z)
        xs = [x for x in xs]
        ys = [y for y in ys]
        
        pcb.add_poly(xs, ys, z=z, name=poly['handle'])
    return pcb
    



from typing import Literal

class Settings:
    
    def __init__(self):
        self.mw_2dbc: bool = True
        self.mw_2dbc_lim: float = 10.0
        self.mw_2dbc_peclim: float = 1e8
        self.mw_3d_peclim: float = 1e7
        
DEFAULT_SETTINGS = Settings()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.
from __future__ import annotations
import numpy as np
from typing import Callable
import inspect
from .const import C0


def num_args(func):
    sig = inspect.signature(func)
    return sum(
        1
        for p in sig.parameters.values()
        if p.default is inspect._empty and p.kind in (
            p.POSITIONAL_ONLY,
            p.POSITIONAL_OR_KEYWORD,
            p.KEYWORD_ONLY
        )
    )
    
def _to_mat(value: float | complex | int | np.ndarray) -> np.ndarray:
    if np.isscalar(value):
        return np.eye(3)*value
    if value.shape in ((3,), (3,1), (1,3)):
        return np.diag(np.ravel(value))
    if value.shape == (3,3):
        return value
    else:
        return ValueError(f'Trying to parse {value} as a material property tensor but it cant be identified as scalar, vector or matrix')

class MatProperty:
    _freq_dependent: bool = False
    _coord_dependent: bool = False
    _pickle_exclude = {"_func","_fmax"}
    """The MatProperty class is an interface for EMerge to deal with frequency and coordinate dependent material properties
    """
    
    def __init__(self, value: float | complex | int | np.ndarray):
        self.value: np.ndarray = _to_mat(value)
        
        self._apply_to: np.ndarray = np.array([], dtype=np.int64)
        self._x: np.ndarray = np.array([], dtype=np.float64)
        self._y: np.ndarray = np.array([], dtype=np.float64)
        self._z: np.ndarray = np.array([], dtype=np.float64)
        
        self._fmax = lambda f: value

    def initialize(self, x: np.ndarray, y: np.ndarray, z: np.ndarray, ids: np.ndarray) -> None:
        self._apply_to = np.concatenate([self._apply_to, ids])
        self._x = np.concatenate([self._x, x])
        self._y = np.concatenate([self._y, y])
        self._z = np.concatenate([self._z, z])
        
    def __call__(self, f: float, data: np.ndarray) -> np.ndarray:
        data[:,:,self._apply_to] = np.repeat(self.value[:,:,np.newaxis], self._apply_to.shape[0], axis=2)
        return data

    def scalar(self, f: float):
        return self.value[0,0]

    def reset(self) -> None:
        self._apply_to: np.ndarray = np.array([], dtype=np.int64)
        self._x: np.ndarray = np.array([], dtype=np.float64)
        self._y: np.ndarray = np.array([], dtype=np.float64)
        self._z: np.ndarray = np.array([], dtype=np.float64)
    
    def __getstate__(self):
        state = self.__dict__.copy()
        for k in self._pickle_exclude:
            state.pop(k, None)
        
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)
        for k in self._pickle_exclude:
            setattr(self, k, None)
    
class FreqDependent(MatProperty):
    _freq_dependent: bool = True
    _coord_dependent: bool = False

    def __init__(self, 
                 scalar: Callable | None = None,
                 vector: Callable | None = None,
                 matrix: Callable | None = None):
        """Creates a frequency dependent property object.
        
        If the property is defined as a scalar value, use the "scalar" argument
        If the property is a diagonal rank-2 tensor, use the "vector" argument                   
        If the property is a full rank-2 tensor, use the "matrix" argument

        The max_value property must be set to tell EMerge how height this value can get 
        as it will be used to define the discretization of the mesh.
        
        Args:
            scalar (Callable | None, optional): The scalar value function returning a float/complex. Defaults to None.
            vector (Callable | None, optional): The diagonal rank-2 tensor function returning a (3,) array. Defaults to None.
            matrix (Callable | None, optional): The rank-2 tensor function returning a (3,3) array. Defaults to None.

        Returns:
            _type_: _description_
        """
        if scalar is not None:
            def _func(f: float) -> np.ndarray:
                return np.eye(3)*scalar(f)
        if vector is not None:
            def _func(f: float) -> np.ndarray:
                return np.diag(np.ravel(vector(f)))
        
        if matrix is not None:
            _func = matrix

        self._func: Callable = _func
        
        self._apply_to: np.ndarray = np.array([], dtype=np.int64)
        self._x: np.ndarray = np.array([], dtype=np.float64)
        self._y: np.ndarray = np.array([], dtype=np.float64)
        self._z: np.ndarray = np.array([], dtype=np.float64)
        
        self._fmax: Callable = lambda f: np.max(np.ravel(self._func(f)))

    def initialize(self, x: np.ndarray, y: np.ndarray, z: np.ndarray, ids: np.ndarray) -> None:
        self._apply_to = np.concatenate([self._apply_to, ids])
        
    def __call__(self, f: float, data: np.ndarray) -> np.ndarray:
        data[:,:,self._apply_to] = np.repeat(self._func(f)[:,:,np.newaxis], self._apply_to.shape[0], axis=2)
        return data
    
    def scalar(self, f: float):
        return self._func(f)[0,0]
    
class CoordDependent(MatProperty):
    _freq_dependent: bool = False
    _coord_dependent: bool = True
    def __init__(self, 
                 max_value: float,
                 scalar: Callable | None = None,
                 vector: Callable | None = None,
                 matrix: Callable | None = None,
                 ):
        """Creates a coordinate dependent property object.
        
        If the property is defined as a scalar value, use the "scalar" argument.
        
        If the property is a diagonal rank-2 tensor, use the "vector" argument.
        
        If the property is a full rank-2 tensor, use the "matrix" argument.
        

        The max_value property must be set to tell EMerge how height this value can get 
        as it will be used to define the discretization of the mesh.
        
        Args:
            max_value (float): The heighest value of the material property
            scalar (Callable | None, optional): The scalar value function returning a float/complex. Defaults to None.
            vector (Callable | None, optional): The diagonal rank-2 tensor function returning a (3,) array. Defaults to None.
            matrix (Callable | None, optional): The rank-2 tensor function returning a (3,3) array. Defaults to None.

        Returns:
            _type_: _description_
        """
        
        if scalar is not None:
            def _func(x, y, z) -> np.ndarray:
                return np.eye(3)[:, :, None] * scalar(x,y,z)[None, None, :]
        
        if vector is not None:
            def _func(x, y, z) -> np.ndarray:
                N = x.shape[0]
                out = np.zeros((3, 3, N), dtype=vector(0,0,0).dtype)
                idx = np.arange(3)
                out[idx, idx, :] = vector(x,y,z)
                return out
        if matrix is not None:
            _func = matrix

        self._func: Callable = _func
        self._apply_to: np.ndarray = np.array([], dtype=np.int64)
        self._x: np.ndarray = np.array([], dtype=np.float64)
        self._y: np.ndarray = np.array([], dtype=np.float64)
        self._z: np.ndarray = np.array([], dtype=np.float64)
        
        self._values: np.ndarray = None
        self._fmax: Callable = lambda f: max_value

    def initialize(self, x: np.ndarray, y: np.ndarray, z: np.ndarray, ids: np.ndarray) -> None:
        self._apply_to = np.concatenate([self._apply_to, ids])
        self._x = np.concatenate([self._x, x])
        self._y = np.concatenate([self._y, y])
        self._z = np.concatenate([self._z, z])
        
    def __call__(self, f: float, data: np.ndarray) -> np.ndarray:
        data[:,:,self._apply_to] = self._func(self._x, self._y, self._z)
        return data

    def scalar(self, f: float):
        return self._func(0,0,0)[0,0]
    
class FreqCoordDependent(MatProperty):
    _freq_dependent: bool = True
    _coord_dependent: bool = True
    
    def __init__(self, 
                 max_value: float,
                 scalar: Callable | None = None,
                 vector: Callable | None = None,
                 matrix: Callable | None = None):
        """Creates a frequency and coordinate dependent property object.
        
        If the property is defined as a scalar value, use the "scalar" argument.
        
        If the property is a diagonal rank-2 tensor, use the "vector" argument.
        
        If the property is a full rank-2 tensor, use the "matrix" argument.

        The max_value property must be set to tell EMerge how height this value can get 
        as it will be used to define the discretization of the mesh.
        
        Args:
            max_value (float): The heighest value of the material property
            scalar (Callable | None, optional): The scalar value function returning a float/complex. Defaults to None.
            vector (Callable | None, optional): The diagonal rank-2 tensor function returning a (3,) array. Defaults to None.
            matrix (Callable | None, optional): The rank-2 tensor function returning a (3,3) array. Defaults to None.

        Returns:
            _type_: _description_
        """
        if scalar is not None:
            def _func(f, x, y, z) -> np.ndarray:
                return np.eye(3)[:, :, None] * scalar(f,x,y,z)[None, None, :]
        if vector is not None:
            def _func(f,x, y, z) -> np.ndarray:
                N = x.shape[0]
                out = np.zeros((3, 3, N), dtype=vector(1e9,0,0,0).dtype)
                idx = np.arange(3)
                out[idx, idx, :] = vector(f,x,y,z)
                return out
        if matrix is not None:
            _func = matrix

        self._func: Callable = _func
        
        self._apply_to: np.ndarray = np.array([], dtype=np.int64)
        self._x: np.ndarray = np.array([], dtype=np.float64)
        self._y: np.ndarray = np.array([], dtype=np.float64)
        self._z: np.ndarray = np.array([], dtype=np.float64)
        
        self._fmax: Callable = lambda f: max_value

    def initialize(self, x: np.ndarray, y: np.ndarray, z: np.ndarray, ids: np.ndarray) -> None:
        self._apply_to = np.concatenate([self._apply_to, ids])
        self._x = np.concatenate([self._x, x])
        self._y = np.concatenate([self._y, y])
        self._z = np.concatenate([self._z, z])
        
    def __call__(self, f: float, data: np.ndarray) -> np.ndarray:
        data[:,:,self._apply_to] = self._func(f,self._x, self._y,self._z)
        return data
    
    def scalar(self, f: float):
        return self._func(f, 0,0,0)[0,0]

class Material:
    """The Material class generalizes a material in the EMerge FEM environment.

    If a scalar value is provided for the relative permittivity or the relative permeability
    it will be used as multiplication entries for the material property diadic as identity matrix.

    Additionally, a frequency, coordinate or both frequency and coordinate dependent material property
    may be supplied for the properties: er, ur, tand and cond.
    
    To supply a frequency-dependent property use: emerge.FreqDependent()
    To supply a coordinate-dependent property use: emerge.CoordDependent()
    to supply a frequency and coordinate dependent property use: emerge.FreqCoordDependent()
    
    """
    _pickle_exclude = {"_neff"}
    
    def __init__(self,
                 er: float | complex | np.ndarray | MatProperty = 1.0,
                 ur: float | complex | np.ndarray | MatProperty = 1.0,
                 tand: float | MatProperty = 0.0,
                 cond: float | MatProperty = 0.0,
                 _neff: float | None = None,
                 color: str ="#BEBEBE",
                 opacity: float = 1.0,
                 _metal: bool = False,
                 name: str = 'unnamed'):
        
        if not isinstance(er, MatProperty):
            er = MatProperty(er)
        if not isinstance(ur, MatProperty):
            ur = MatProperty(ur)
        if not isinstance(tand, MatProperty):
            tand = MatProperty(tand)
        if not isinstance(cond, MatProperty):
            cond = MatProperty(cond)
        
        self.name: str = name
        self.er: MatProperty = er
        self.ur: MatProperty = ur
        self.tand: MatProperty = tand
        self.cond: MatProperty = cond

        self.color: str = color
        self.opacity: float = opacity
        self._hash_key: int = -1
        
        if _neff is None:
            self._neff: Callable = lambda f: np.sqrt(self.ur._fmax(f)*self.er._fmax(f))
        else:
            self._neff: Callable = lambda f: _neff
        hex_str = self.color.lstrip('#')
        self._color_rgb = tuple(int(hex_str[i:i+2], 16)/255.0 for i in (0, 2, 4))
        self._metal: bool = _metal
    
    def __getstate__(self):
        state = self.__dict__.copy()
        for k in self._pickle_exclude:
            state.pop(k, None)
        
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)
        for k in self._pickle_exclude:
            setattr(self, k, None)

    def __hash__(self):
        return self._hash_key
    
    def __str__(self) -> str:
        return f'Material({self.name}, {self._hash_key})'
    
    def __repr__(self):
        return f'Material({self.name}, {self._hash_key})'
    
    def initialize(self, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray, ids: np.ndarray):
        """Initializes the Material properties to be evaluated at xyz-coordinates for
        a given set of tetrahedral ids.

        Args:
            xs (np.ndarray): The tet-centroid x-coordinates
            ys (np.ndarray): The tet-centroid y-coordinates
            zs (np.ndarray): The tet-centroid z-coordinates
            ids (np.ndarray): The tet-indices
        """
        self.er.initialize(xs, ys, zs, ids)
        self.ur.initialize(xs, ys, zs, ids)
        self.tand.initialize(xs, ys, zs, ids)
        self.cond.initialize(xs, ys, zs, ids)
    
    def reset(self) -> None:
        """Resets assignment of material properties to coordiantes and tetrahedral indices.
        """
        self.er.reset()
        self.ur.reset()
        self.tand.reset()
        self.cond.reset()
        self._hash_key = -1
        
    @property
    def frequency_dependent(self) -> bool:
        """If The material property are at all frequency dependent.

        Returns:
            bool: _description_
        """
        return self.er._freq_dependent or self.ur._freq_dependent or self.tand._freq_dependent or self.cond._freq_dependent
    
    @property
    def coordinate_dependent(self) -> bool:
        """If the material properties are at all coordinate dependent

        Returns:
            bool: _description_
        """
        return self.er._coord_dependent or self.ur._coord_dependent or self.tand._coord_dependent or self.cond._coord_dependent
    
    
    def neff(self, f: float):
        """ Computes the maximum occuring effective refractive index for this material."""
        return self._neff(f)
    
    @property
    def color_rgb(self) -> tuple[float,float,float]:
        return self._color_rgb
     
    @staticmethod
    def drude_model(conductivity: float, 
                    colission_time: float, 
                    er: float = 1.0,
                    ur: float = 1.0, 
                    color: str = "#aaaaaa", 
                    opacity: float = 0.3,
                    metal: bool = True) -> Material:
        """Creates a Material using the Drume model for conductivity
        Requires at least the DC bulk condutivity σ₀ [S/m] and the
        collision time τ.

        Args:
            conductivity (float): The DC bulk conductivity σ₀ in S/m
            colission_time (float): The collision time.
            er (float, optional): The dielectric constant. Defaults to 1.0.
            ur (float, optional): The relative permeability. Defaults to 1.0.
            color (str, optional): The material rendering color. Defaults to "#aaaaaa".
            opacity (float, optional): The material rendering opacity. Defaults to 0.3.
            metal (bool, optional): If it should be rendered as a metal.. Defaults to True.

        Returns:
            Material: The resultant material.
        """
        colission_dist = colission_time/C0
        fsigma = FreqDependent(scalar = lambda f: conductivity/(1 - 1j*2*np.pi*f*colission_dist))
        return Material(er, ur, 0.0, fsigma, color=color, opacity=opacity, _metal=metal)
    
AIR = Material(color="#4496f3", opacity=0.05, name='Air')
COPPER = Material(cond=5.8e7, color="#62290c", _metal=True, name='Copper')
PEC = Material(color="#ff78aa", opacity=1.0, cond=1e30, _metal=True, name="PEC")


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from numba.core import event
from numba import njit

_COMPILE_MESSAGE = """
 [ EMERGE ]
⚠  Numba will be compiling optimized code in this first run; this may take a few minutes.
   • Additional functions may be compiled on-the-fly.
   • Compilation happens only once-subsequent runs load from cache.
   Please wait…"""

@njit(cache=True)
def _donothing(a):
    return a

class Notify(event.Listener):
    def on_start(self, ev):
        f = ev.data['dispatcher']
        sig = ev.data['args']
        if f is _donothing:          # limit to the function you care about
            sig = ev.data['args']
            print(_COMPILE_MESSAGE)

    def on_end(self, ev):        # unused here
        pass


# install listener only for this block:
with event.install_listener("numba:compile", Notify()):
    _donothing(0) 


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import gmsh # type: ignore
import numpy as np
from .cs import Axis, CoordinateSystem, _parse_vector, Plane
from typing import Callable, TypeVar, Iterable, Any

def align_rectangle_frame(pts3d: np.ndarray, normal: np.ndarray) -> dict[str, Any]:
    """Tries to find a rectangle as convex-hull of a set of points with a given normal vector.

    Args:
        pts3d (np.ndarray): The points (N,3)
        normal (np.ndarray): The normal vector.

    Returns:
        dict[str, np.ndarray]: The output data
    """
    
    # 1. centroid
    from scipy.spatial import ConvexHull # type: ignore
    
    Omat = np.squeeze(np.mean(pts3d, axis=0))

    # 2. build e_x, e_y
    n = np.squeeze(normal/np.linalg.norm(normal))
    seed = np.array([1.,0.,0.])
    if abs(seed.dot(n)) > 0.9:
        seed = np.array([0.,1.,0.])
    e_x = seed - n*(seed.dot(n))
    e_x /= np.linalg.norm(e_x)
    e_y = np.cross(n, e_x)

    # 3. project into 2D
    pts2d = np.vstack([[(p-Omat).dot(e_x), (p-Omat).dot(e_y)] for p in pts3d])

    # 4. convex hull
    hull = ConvexHull(pts2d)
    hull_pts = pts2d[hull.vertices]

    # 5. rotating calipers: find min-area rectangle
    best = (None, np.inf, None)  # (angle, area, (xmin,xmax,ymin,ymax))
    for i in range(len(hull_pts)):
        p0 = hull_pts[i]
        p1 = hull_pts[(i+1)%len(hull_pts)]
        edge = p1 - p0
        theta = -np.arctan2(edge[1], edge[0])  # rotate so edge aligns with +X
        R = np.array([[np.cos(theta), -np.sin(theta)],
                      [np.sin(theta),  np.cos(theta)]])
        rot = hull_pts.dot(R.T)
        xmin, ymin = rot.min(axis=0)
        xmax, ymax = rot.max(axis=0)
        area = (xmax-xmin)*(ymax-ymin)
        if area < best[1]:
            best = (theta, area, (xmin,xmax,ymin,ymax), R) # type: ignore

    theta, _, (xmin,xmax,ymin,ymax), R = best # type: ignore

    # 6. rectangle axes in 3D
    u =  np.cos(-theta)*e_x + np.sin(-theta)*e_y
    v = -np.sin(-theta)*e_x + np.cos(-theta)*e_y

    # corner points in 3D:
    corners = []
    for a in (xmin, xmax):
        for b in (ymin, ymax):
            # back-project to the original 2D frame:
            p2 = np.array([a, b]).dot(R)  # rotate back
            P3 = Omat + p2[0]*e_x + p2[1]*e_y
            corners.append(P3)

    return {
      "origin": Omat,
      "axes": (u, v, n),
      "corners": np.array(corners).reshape(4,3)
    }

TSelection = TypeVar("TSelection", bound="Selection")

# Your custom alphabet
ALPHABET = (
    "abcdefghijklmnopqrstuvwxyz"
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "012345%#"
)
# Map char → int and int → char
CHAR_TO_VAL = {ch: i for i, ch in enumerate(ALPHABET)}
VAL_TO_CHAR = {i: ch for i, ch in enumerate(ALPHABET)}

def encode_data(values: tuple[float,...]) -> str:
    """
    Convert a tuple of floats into a custom base64-like encoded string
    using 16-bit float representation and a 64-character alphabet.
    """
    # Convert floats to 16-bit half-floats
    arr = np.array(values, dtype=np.float16)
    # Get raw bytes
    byte_data = arr.tobytes()
    # Convert bytes to a bitstring
    bitstring = ""
    for byte in byte_data:
        bitstring += f"{byte:08b}"
    
    # Pad the bitstring to a multiple of 6 bits
    pad_len = (6 - len(bitstring) % 6) % 6
    bitstring += "0" * pad_len
    
    # Encode 6 bits at a time
    encoded = ""
    for i in range(0, len(bitstring), 6):
        chunk = bitstring[i:i+6]
        val = int(chunk, 2)
        encoded += VAL_TO_CHAR[val]
    
    # Optionally store how many pad bits we added
    # so we can remove them during decoding.
    # We'll prepend it as one character encoding pad_len (0-5).
    encoded = VAL_TO_CHAR[pad_len] + encoded
    return encoded

def decode_data(encoded: str) -> tuple[float,...]:
    """
    Decode a string produced by floats_to_custom_string
    back into a tuple of float64 values.
    """
    # The first character encodes how many zero bits were padded
    pad_char = encoded[0]
    pad_len = CHAR_TO_VAL[pad_char]
    data = encoded[1:]
    
    # Convert each char back to 6 bits
    bitstring = ""
    for ch in data:
        val = CHAR_TO_VAL[ch]
        bitstring += f"{val:06b}"
    
    # Remove any padding bits at the end
    if pad_len > 0:
        bitstring = bitstring[:-pad_len]
    
    # Split into bytes
    byte_data = []
    for i in range(0, len(bitstring), 8):
        byte_chunk = bitstring[i:i+8]
        if len(byte_chunk) < 8:
            break
        byte_val = int(byte_chunk, 2)
        byte_data.append(byte_val)
    
    byte_array = bytes(byte_data)
    # Recover as float16
    arr = np.frombuffer(byte_array, dtype=np.float16)
    # Convert back to float64 for higher precision
    return tuple(np.array(arr, dtype=np.float64))

class Selection:
    """A generalized class representing a slection of tags.

    """
    dim: int = -1
    def __init__(self, tags: list[int] | set[int] | tuple[int] | None = None):

        self._tags: set[int] = set()

        if tags is not None:
            if not isinstance(tags, (list,set,tuple)):
                raise TypeError(f'Argument tags must be of type list, tuple or set, instead its {type(tags)}')
            self._tags = set(tags)
    
    @staticmethod
    def from_dim_tags(dim: int, tags: list[int] | set[int]) -> Selection:
        if dim==0:
            return PointSelection(tags)
        elif dim==1:
            return EdgeSelection(tags)
        elif dim==2:
            return FaceSelection(tags)
        elif dim==3:
            return DomainSelection(tags)
        raise ValueError(f'Dimension must be 0,1,2 or 3. Not {dim}')
    
    @property
    def tags(self) -> list[int]:
        return list(self._tags)
    
    @property
    def color_rgb(self) -> tuple[float, float, float]:
        return (0.5,0.5,1.0)
    
    @property
    def centers(self) -> list[tuple[float, float, float],]:
        return [gmsh.model.occ.get_center_of_mass(self.dim, tag) for tag in self.tags]
    
    @property
    def _metal(self) -> bool:
        return False
    
    @property
    def opacity(self) -> float:
        return 0.6
    ####### DUNDER METHODS
    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.tags})'
    
    ####### PROPERTIES
    @property
    def dimtags(self) -> list[tuple[int,int]]:
        return [(self.dim, tag) for tag in self.tags]
    
    @property
    def center(self) -> np.ndarray | list[np.ndarray]:
        if len(self.tags)==1:
            return gmsh.model.occ.getCenterOfMass(self.dim, self.tags[0])
        else:
            return [gmsh.model.occ.getCenterOfMass(self.dim, tag) for tag in self.tags]
    
    @property
    def points(self) -> list[np.ndarray]:
        '''A list of 3D coordinates of all nodes comprising the selection.'''
        points = gmsh.model.get_boundary(self.dimtags, recursive=True)
        coordinates = [gmsh.model.getValue(*p, []) for p in points]
        return coordinates
    
    @property
    def bounding_box(self) -> tuple[Iterable, Iterable]:
        if len(self.tags)==1:
            return gmsh.model.occ.getBoundingBox(self.dim, self.tags[0])
        else:
            minx = miny = minz = 1e10
            maxx = maxy = maxz = -1e10
            for tag in self.tags:
                x0, y0, z0, x1, y1, z1 = gmsh.model.occ.getBoundingBox(self.dim, tag)
                minx = min(minx, x0)
                miny = min(miny, y0)
                minz = min(minz, z0)
                maxx = max(maxx, x1)
                maxy = max(maxy, y1)
                maxz = max(maxz, z1)
            return (minx, miny, minz), (maxx, maxy, maxz)
        
    def exclude(self, xyz_excl_function: Callable = lambda x,y,z: True, plane: Plane | None = None, axis: Axis | None = None) -> Selection:
        """Exclude points by evaluating a function(x,y,z)-> bool

        This modifies the selection such that the selection does not contain elements
        of this selection of which the center of mass is excluded by the exclusion function.

        Args:
            xyz_excl_function (Callable): A callable for (x,y,z) that returns True if the point should be excluded.

        Returns:
            Selection: This Selection modified without the excluded points.
        """
        include = [~xyz_excl_function(*gmsh.model.occ.getCenterOfMass(*tag)) for tag in self.dimtags]
        
        if axis is not None:
            norm = axis.np
            include2 = [abs(gmsh.model.getNormal(tag, np.array([0,0]))@norm)<0.9 for tag in self.tags]
            include = [i1 for i1, i2 in zip(include, include2) if i1 and i2]
        self._tags = set([t for incl, t in zip(include, self._tags) if incl])
        return self
    
    def isolate(self, xyz_excl_function: Callable = lambda x,y,z: True, plane: Plane | None = None, axis: Axis | None = None) -> Selection:
        """Include points by evaluating a function(x,y,z)-> bool

        This modifies the selection such that the selection does not contain elements
        of this selection of which the center of mass is excluded by the exclusion function.

        Args:
            xyz_excl_function (Callable): A callable for (x,y,z) that returns True if the point should be excluded.

        Returns:
            Selection: This Selection modified without the excluded points.
        """
        include1 = [xyz_excl_function(*gmsh.model.occ.getCenterOfMass(*tag)) for tag in self.dimtags]
        
        if axis is not None:
            norm = axis.np
            include2 = [(gmsh.model.getNormal(tag, np.array([0,0]))@norm)>0.99 for tag in self.tags]
            include1 = [i1 for i1, i2 in zip(include1, include2) if i1 and i2]
        self._tags = set([t for incl, t in zip(include1, self._tags) if incl])
        return self

    def __operable__(self, other: Selection) -> None:
        if not self.dim == other.dim:
            raise ValueError(f'Selection dimensions must be equal. Trying to operate on dim {self.dim} and {other.dim}')
        pass

    def __add__(self, other: Selection) -> Selection:
        self.__operable__(other)
        return Selection.from_dim_tags(self.dim, self._tags.union(other._tags))
    
    def __and__(self, other: Selection) -> Selection:
        self.__operable__(other)
        return Selection.from_dim_tags(self.dim, self._tags.intersection(other._tags))
    
    def __or__(self, other: Selection) -> Selection:
        self.__operable__(other)
        return Selection.from_dim_tags(self.dim, self._tags.union(other._tags))
    
    def __sub__(self, other: Selection) -> Selection:
        self.__operable__(other)
        return Selection.from_dim_tags(self.dim, self._tags.difference(other.tags))

class PointSelection(Selection):
    """A Class representing a selection of points.

    """
    dim: int = 0
    def __init__(self, tags: list[int] | set [int] | None = None):
        super().__init__(tags)

class EdgeSelection(Selection):
    """A Class representing a selection of edges.

    """
    dim: int = 1
    def __init__(self, tags: list[int] | set [int] | None = None):
        super().__init__(tags)

class FaceSelection(Selection):
    """A Class representing a selection of Faces.

    """
    dim: int = 2
    def __init__(self, tags: list[int] | set [int] | None = None):
        super().__init__(tags)

    @property
    def normal(self) -> np.ndarray:
        ''' Returns a 3x3 coordinate matrix of the XY + out of plane basis matrix defining the face assuming it can be projected on a flat plane.'''
        ns = [gmsh.model.getNormal(tag, np.array([0,0])) for tag in self.tags]
        return ns[0]
    
    def rect_basis(self) -> tuple[CoordinateSystem, tuple[float, float]]:
        ''' Returns a dictionary with keys: origin, axes, corners. The axes are the 3D basis vectors of the rectangle. The corners are the 4 corners of the rectangle.
        
        Returns:
            cs: CoordinateSystem: The coordinate system of the rectangle.
            size: tuple[float, float]: The size of the rectangle (width [m], height[m])
        '''
        if len(self.tags) != 1:
            raise ValueError('rect_basis only works for single face selections')
        
        pts3d = self.points
        normal = self.normal
        data = align_rectangle_frame(np.array(pts3d), normal)
        plane = data['axes'][:2]
        origin = data['origin']

        cs = CoordinateSystem(Axis(plane[0]), Axis(plane[1]), Axis(data['axes'][2]), origin)

        size1 = float(np.linalg.norm(data['corners'][1] - data['corners'][0]))
        size2 = float(np.linalg.norm(data['corners'][2] - data['corners'][0]))

        if size1>size2:
            cs.swapxy()
            return cs, (size1, size2)
        else:
            return cs, (size2, size1)
            
    def sample(self, Npts: int) -> tuple[np.ndarray, np.ndarray, np.ndarray] | list[tuple[np.ndarray, np.ndarray, np.ndarray]]:
        ''' Sample the surface at the compiler defined parametric coordinate range.
        This function usually returns a square region that contains the surface.
        
        Returns:
        --------
            X: np.ndarray
                a NxN numpy array of X coordinates.
            Y: np.ndarray
                a NxN numpy array of Y coordinates.
            Z: np.ndarray
                a NxN numpy array of Z coordinates'''
        coordset: list[tuple[np.ndarray, np.ndarray, np.ndarray]] = []
        for tag in self.tags:
            tags, coord, param = gmsh.model.mesh.getNodes(2, tag, includeBoundary=True)
            
            uss = param[0::2]
            vss = param[1::2]

            umin = min(uss)
            umax = max(uss)
            vmin = min(vss)
            vmax = max(vss)

            us = np.linspace(umin, umax, Npts)
            vs = np.linspace(vmin, vmax, Npts)

            U, V = np.meshgrid(us, vs, indexing='ij')

            shp = U.shape

            uax = U.flatten()
            vax = V.flatten()
            
            pcoords = np.zeros((2*uax.shape[0],))

            pcoords[0::2] = uax
            pcoords[1::2] = vax

            coords = gmsh.model.getValue(2, tag, pcoords).reshape(-1,3).T
           
            coordset.append((coords[0,:].reshape(shp), 
                             coords[1,:].reshape(shp), 
                             coords[2,:].reshape(shp)))
            
        X = [c[0] for c in coordset]
        Y = [c[1] for c in coordset]
        Z = [c[2] for c in coordset]
        if len(X) == 1:
            return X[0], Y[0], Z[0]
        return np.array(X), np.array(Y), np.array(Z)
    
class DomainSelection(Selection):
    """A Class representing a selection of domains.

    """
    dim: int = 3
    def __init__(self, tags: list[int] | set[int] | None = None):
        super().__init__(tags)

SELECT_CLASS: dict[int, type[Selection]] = {
    0: PointSelection,
    1: EdgeSelection,
    2: FaceSelection,
    3: DomainSelection
}

######## SELECTOR

class Selector:
    """A class instance with convenient methods to generate selections using method chaining.

    Use the specific properties and functions in a "language" like way to make selections.

    To specify what to select, use the .node, .edge, .face or .domain property.
    These properties return the Selector after which you can say how to execute a selection.


    """
    def __init__(self):
        self._current_dim: int = -1
    
    ## DIMENSION CHAIN
    @property
    def node(self) -> Selector:
        self._current_dim = 0
        return self

    @property
    def edge(self) -> Selector:
        self._current_dim = 1
        return self
    
    @property
    def face(self) -> Selector:
        self._current_dim = 2
        return self

    @property
    def domain(self) -> Selector:
        self._current_dim = 3
        return self
    
    def near(self,
             x: float,
             y: float,
             z: float = 0) -> Selection | PointSelection | EdgeSelection | FaceSelection | DomainSelection:
        """Returns a selection of the releative dimeions by which of the instances is most proximate to a coordinate.

        Args:
            x (float): The X-coordinate
            y (float): The Y-coordinate
            z (float, optional): The Z-coordinate. Defaults to 0.

        Returns:
            Selection | PointSelection | EdgeSelection | FaceSelection | DomainSelection: The resultant selection.
        """
        dimtags = gmsh.model.getEntities(self._current_dim)
        
        
        dists = [np.linalg.norm(np.array([x,y,z]) - gmsh.model.occ.getCenterOfMass(*tag)) for tag in dimtags]
        index_of_closest = np.argmin(dists)

        return SELECT_CLASS[self._current_dim]([dimtags[index_of_closest][1],])
    
    def inlayer(self, 
                x: float,
                y: float,
                z: float,
                vector: tuple[float, float, float] | np.ndarray | Axis) -> Selection:
        '''Returns a list of selections that are in the layer defined by the plane normal vector and the point (x,y,z)
        
        The layer is specified by two infinite planes normal to the provided vector. The first plane is originated
        at the provided origin. The second plane is placed such that it contains the point origin+vector.

        Args:
            x (float): The X-coordinate from which to select
            y (float): The Y-coordinate from which to select
            z (float): The Z-coordinate from which to select
            vector (np.ndarray, tuple, Axis): A vector with length in (meters) originating at the origin.

        Returns:
            Selection | PointSelection | EdgeSelection | FaceSelection | DomainSelection: The resultant selection.

        '''
        vector = _parse_vector(vector)
        
        dimtags = gmsh.model.getEntities(self._current_dim)

        coords = [gmsh.model.occ.getCenterOfMass(*tag) for tag in dimtags]

        L = np.linalg.norm(vector)
        vector = vector / L

        output = []
        for i, c in enumerate(coords):
            c_local = c - np.array([x,y,z])
            if 0 < np.dot(vector, c_local) < L: # type: ignore
                output.append(dimtags[i][1])
        return SELECT_CLASS[self._current_dim](output)
    
    def inplane(self,
                x: float,
                y: float,
                z: float,
                normal_axis: Axis | tuple[float, float, float] | None = None,
                plane: Plane | None = None,
                tolerance: float = 1e-8) -> FaceSelection:
        """Returns a FaceSelection for all faces that lie in a provided infinite plane
        specified by an origin plus a plane normal vector.

        Args:
            x (float): The plane origin X-coordinate
            y (float): The plane origin Y-coordinate
            z (float): The plane origin Z-coordinate
            normal_axis (Axis, tuple): The plane normal vector
            tolerance (float, optional): An in plane tolerance (displacement and normal dot product). Defaults to 1e-6.

        Returns:
            FaceSelection: All faces that lie in the specified plane
        """
        orig = np.array([x,y,z])
        if plane is not None:
            norm = plane.normal.np
        elif normal_axis is not None:
            norm = _parse_vector(normal_axis)
            norm = norm/np.linalg.norm(norm)
        else:
            raise RuntimeError('No plane or axis defined for selection.')
        
        dimtags = gmsh.model.getEntities(2)
        coords = [gmsh.model.occ.getCenterOfMass(*tag) for tag in dimtags]
        normals = [gmsh.model.get_normal(t, (0,0)) for d,t, in dimtags]
        tags = []
        for (d,t), o, n in zip(dimtags, coords, normals):
            normdist = np.abs((o-orig) @ norm)
            dotnorm = np.abs(n @ norm)
            if normdist < tolerance and dotnorm > 1-tolerance:
                tags.append(t)
        return FaceSelection(tags)
    
    def code(self, code: str):
        nums1 = decode_data(code)
        
        dimtags = gmsh.model.getEntities(2)
        
        scoring = dict()
        for dim, tag in dimtags:
            x1, y1, z1 = gmsh.model.occ.getCenterOfMass(2, tag)
            xmin2, ymin2, zmin2, xmax2, ymax2, zmax2 = gmsh.model.occ.getBoundingBox(dim, tag)
            nums2 = [x1, y1, z1, xmin2, ymin2, zmin2, xmax2, ymax2, zmax2]
            score = np.sqrt(sum([(a-b)**2 for a,b in zip(nums1, nums2)]))
            scoring[tag] = score
        
        min_val = min(scoring.values())

        # Find all keys whose value == min_val
        candidates = [k for k, v in scoring.items() if v == min_val]

        # Pick the lowest key
        lowest_key = min(candidates)
        return FaceSelection([lowest_key,])

SELECTOR_OBJ = Selector()




# from ...selection import FaceSelection, SELECTOR_OBJ
# import numpy as np
# from typing import Generator
# from .microwave_bc import Periodic, FloquetPort
# from ...geo.extrude import XYPolygon, GeoPrism
# from ...geo import XYPlate, Alignment
# from ...periodic import PeriodicCell

# class MWPeriodicCell(PeriodicCell):

#     def __post_init__(self):
#         self._bcs: list[Periodic] = []
#         self._ports: list[FloquetPort] = []

#     def volume(self, z1: float, z2: float) -> GeoPrism:
#         """Genereates a volume with the cell geometry ranging from z1 tot z2

#         Args:
#             z1 (float): The start height
#             z2 (float): The end height

#         Returns:
#             GeoPrism: The resultant prism
#         """
#         raise NotImplementedError('This method is not implemented for this subclass.')
    
#     def floquet_port(self, z: float) -> FloquetPort:
#         raise NotImplementedError('This method is not implemented for this subclass.')
    
#     def cell_data(self) -> Generator[tuple[FaceSelection,FaceSelection,tuple[float, float, float]], None, None]:
#         """An iterator that yields the two faces of the hex cell plus a cell periodicity vector

#         Yields:
#             Generator[np.ndarray, np.ndarray, np.ndarray]: The face and periodicity data
#         """
#         raise NotImplementedError('This method is not implemented for this subclass.')

#     @property
#     def bcs(self) -> list[Periodic]:
#         """Returns a list of Periodic boundary conditions for the given PeriodicCell

#         Args:
#             exclude_faces (list[FaceSelection], optional): A possible list of faces to exclude from the bcs. Defaults to None.

#         Returns:
#             list[Periodic]: The list of Periodic boundary conditions
#         """
#         if not self._bcs:
#             bcs = []
#             for f1, f2, a in self.cell_data():
#                 if self.excluded_faces is not None:
#                     f1 = f1 - self.excluded_faces
#                     f2 = f2 - self.excluded_faces
#                 bcs.append(Periodic(f1, f2, a))
#             self._bcs = bcs
#         return self._bcs
    
#     def set_scanangle(self, theta: float, phi: float, degree: bool = True) -> None:
#         """Sets the scanangle for the periodic condition. (0,0) is defined along the Z-axis

#         Args:
#             theta (float): The theta angle
#             phi (float): The phi angle
#             degree (bool): If the angle is in degrees. Defaults to True
#         """
#         if degree:
#             theta = theta*np.pi/180
#             phi = phi*np.pi/180

        
#         ux = np.sin(theta)*np.cos(phi)
#         uy = np.sin(theta)*np.sin(phi)
#         uz = np.cos(theta)
#         for bc in self._bcs:
#             bc.ux = ux
#             bc.uy = uy
#             bc.uz = uz
#         for port in self._ports:
#             port.scan_theta = theta
#             port.scan_phi = phi

        


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Callable
from numba import njit, f8, i8, c16
from .optimized import gaus_quad_tri, generate_int_points_tri, calc_area


############################################################
#                 NUMBA OPTIMIZED INTEGRALS                #
############################################################

@njit(c16(f8[:,:], i8[:,:], c16[:], f8[:,:], c16[:,:]), cache=True, nogil=True)
def _fast_integral_c(nodes, triangles, constants, DPTs, field_values):
    tot = np.complex128(0.0)

    for it in range(triangles.shape[1]):
        vertex_ids = triangles[:, it]
        v1 = nodes[:,vertex_ids[0]]
        v2 = nodes[:,vertex_ids[1]]
        v3 = nodes[:,vertex_ids[2]]
        A = calc_area(v1, v2, v3)
        field = np.sum(DPTs[0,:]*field_values[:,it])
        tot = tot + constants[it] * field * A
    return tot

@njit(f8(f8[:,:], i8[:,:], f8[:], f8[:,:], f8[:,:]), cache=True, nogil=True)
def _fast_integral_f(nodes, triangles, constants, DPTs, field_values):
    tot = np.float64(0.0)
    for it in range(triangles.shape[1]):
        vertex_ids = triangles[:, it]
        v1 = nodes[:,vertex_ids[0]]
        v2 = nodes[:,vertex_ids[1]]
        v3 = nodes[:,vertex_ids[2]]
        A = calc_area(v1, v2, v3)
        field = np.sum(DPTs[0,:]*field_values[:,it])
        tot = tot + constants[it] * field * A
    return tot


############################################################
#                      PYTHON WRAPPER                     #
############################################################

def surface_integral(nodes: np.ndarray, 
                     triangles: np.ndarray, 
                     function: Callable, 
                     constants: np.ndarray | None = None,
                     gq_order: int = 4) -> complex:
    """Computes the surface integral of a scalar-field

    Computes I = Σ ∫∫ C ᐧ f(x,y,z) dA

    Args:
        nodes (np.ndarray): The integration nodes
        triangles (np.ndarray): The integration triangles
        function (Callable): The scalar field f(x,y,z)
        constants (np.ndarray, optional): The constant C. Defaults to None.
        ndpts (int, optional): Order of gauss quadrature points. Defaults to 4.

    Returns:
        complex: The integral I.
    """
    if constants is None:
        constants = np.ones(triangles.shape[1])
        
    DPTs = gaus_quad_tri(gq_order)
    xall_flat, yall_flat, zall_flat, shape = generate_int_points_tri(nodes, triangles, DPTs)

    fvals = function(xall_flat, yall_flat, zall_flat)

    fA = fvals.reshape(shape)

    if np.iscomplexobj(fA) or np.iscomplexobj(constants):
        return _fast_integral_c(nodes, triangles, constants.astype(np.complex128), DPTs, fA.astype(np.complex128))
    else:
        return _fast_integral_f(nodes, triangles, constants, DPTs, fA)
    


from typing import Iterable, Sequence, Optional
import numpy as np
from matplotlib.colors import LinearSegmentedColormap, to_rgba

def make_colormap(
    hex_colors: Sequence[str],
    positions: Optional[Iterable[float]] = None,
    name: str = "custom",
    N: int = 256,
) -> LinearSegmentedColormap:
    """
    Create a Matplotlib colormap from hex colors with optional positions.

    Parameters
    ----------
    hex_colors : sequence of str
        Hex color strings like '#RRGGBB' (or 'RRGGBB'). At least 2 required.
    positions : iterable of float in [0, 1], optional
        Locations of each color along the gradient. If not provided,
        they are evenly spaced via linspace(0, 1, len(hex_colors)).
        If provided, they do not have to be sorted; they will be sorted
        together with the colors. If the first/last position do not
        hit 0 or 1, they’ll be extended by duplicating the end colors.
    name : str
        Name for the colormap.
    N : int
        Resolution (number of lookup entries) of the colormap.

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
    """
    # Normalize hex strings and basic validation
    colors = []
    for c in hex_colors:
        if not isinstance(c, str):
            raise TypeError("All colors must be hex strings.")
        colors.append(c if c.startswith("#") else f"#{c}")

    if len(colors) < 2:
        raise ValueError("Provide at least two hex colors.")

    # Build/validate positions
    if positions is None:
        pos = np.linspace(0.0, 1.0, len(colors), dtype=float)
    else:
        pos = np.asarray(list(positions), dtype=float)
        if pos.size != len(colors):
            raise ValueError("`positions` must have the same length as `hex_colors`.")
        if np.any((pos < 0.0) | (pos > 1.0)):
            raise ValueError("All positions must be within [0, 1].")

        # Sort positions and carry colors along
        order = np.argsort(pos)
        pos = pos[order]
        colors = [colors[i] for i in order]

        # Ensure coverage of [0, 1] by extending ends if needed
        if pos[0] > 0.0:
            pos = np.r_[0.0, pos]
            colors = [colors[0]] + colors
        if pos[-1] < 1.0:
            pos = np.r_[pos, 1.0]
            colors = colors + [colors[-1]]

    # Pair positions with RGBA
    segment = list(zip(pos.tolist(), (to_rgba(c) for c in colors)))

    # Create the colormap
    return LinearSegmentedColormap.from_list(name, segment, N=N)



import emerge as em
from emerge.plot import plot_ff
import numpy as np

""" CONICAL HORN ANTENNA

This demo uses the revolve feature to build a conical horn antenna. The conical
horn antenna is created by revolving a 2D XYPolygon profile. We compute and plot 
the far-field radiation pattern at an operating frequency of 10GHz.

The dimensions come from this video:
https://www.youtube.com/watch?v=DuXdLuYBGQk

Demo by Edvin Berling
"""

# --- Units ---------------------------------------------------------------
cm = 0.01   # meters per centimeter

# --- Horn and feed dimensions -------------------------------------------
aperture_radius  = 10.334/2 * cm    # horn aperture radius
aperture_length  = 7.809 * cm       # horn length

waveguide_radius = 2.779/2 * cm     # feed waveguide radius
waveguide_length = 2.872 * cm       # feed waveguide length

airbox_length = 6*cm                # airbox length
airbox_width = 15*cm                # airbox width

# --- Create simulation object -------------------------------------------
model = em.Simulation('ConicalHornAntenna')
model.check_version("1.0.6") # Checks version compatibility

# --- Feed geometry -------------------------------------------------------
feed = em.geo.Cylinder(
    waveguide_radius, 
    waveguide_length, 
    cs=em.YZPLANE.cs()
)
# The fundamental mode of a circular waveguide has two version that are 90 degrees rotated due to the circular
# Symmetry of the wavegudie. The two modes can have any arbitrary rotation. This usually depends on slight mesh
# errors. It is  possible to "force" the fundamental mode to be aligned along the Z-axis.

# Stretching the waveguide by 0.1% in the Y-direction pushes the two orthogonal modes slightly apart
# causing one to be perfectly aligned in the Z-drection and the other at a slightly larger propagation constant
# in the Y-direction.
em.geo.stretch(feed, fy=1.001) 

# --- Horn geometry (revolved polygon) -----------------------------------
# Define polygon profile: (x = length, y = radius)
horn_poly = em.geo.XYPolygon(
    [waveguide_length, aperture_length+waveguide_length, aperture_length+waveguide_length, waveguide_length],
    [0, 0, aperture_radius, waveguide_radius]
)
# Revolve polygon around X-axis to create 3D horn
horn_vol = horn_poly.revolve(em.XZPLANE.cs(), (0,0,0), (1,0,0))

# --- Surrounding air --------------------------------------------
air = em.geo.Box(airbox_length, airbox_width, airbox_width, 
                 (aperture_length+waveguide_length,-airbox_width/2,-airbox_width/2))

# --- Finalize geometry --------------------------------------------------
model.commit_geometry()

# --- Solver setup -------------------------------------------------------
model.mw.set_frequency(10e9)                  # 10GHz frequency
model.mw.set_resolution(0.24)                  # mesh resolution fraction

model.generate_mesh()
model.view(selections=[feed.front,])
model.view(selections=[horn_vol.sides,])
model.view(selections=[air.face(no='left')], plot_mesh=True)

# --- Boundary conditions ------------------------------------------------
port1 = model.mw.bc.ModalPort(feed.front, 1)  # excite port at waveguide
radiation_boundary = air.boundary(exclude="left") # open faces
abc = model.mw.bc.AbsorbingBoundary(radiation_boundary)

# --- Run frequency-domain solver ----------------------------------------
data = model.mw.run_sweep()

# --- Far-field radiation pattern (2D cut) -------------------------------
ff_data = data.field[0].farfield_2d(
    (1, 0, 0), (0, 1, 0), radiation_boundary,
    (-90, 90))
plot_ff(ff_data.ang * 180/np.pi, ff_data.normE/em.lib.EISO, dB=True, ylabel='Gain [dBi]')

# --- Visualization ------------------------------------------------------
model.display.add_object(horn_vol, opacity=0.1)
model.display.add_object(feed, opacity=0.1)
model.display.add_surf(
    *data.field[0].farfield_3d(radiation_boundary).surfplot(
        'normE', 'abs', True, True, -10, 5*cm, (waveguide_length+aperture_length,0,0)
    ),
)
model.display.add_portmode(port1, k0=data.field[0].k0)
model.display.show()


import emerge as em
import numpy as np
from emerge.plot import plot_ff_polar, plot_sp

"""
In this demonstration we are going to show how one can make selections using the 
available tools in EMerge. 

To understand the need for the given selection methods it is important to understand
what the difficulty is.

Any CAD implementation will create objects/shapes that can be subtracted or added
from one another. The direct result of this is that any face that logically exists
or is definable (like the front) on an input geometry (of say a box) is not guaranteed 
to exist after manipulations of the shapes. Additionally, the face tags that GMSH keeps
track off will change after you make changes to the geometries.

To still facilitate face selection, EMerge implements a heuristic face
selection algorithm that defines faces of objects as  by two sets of data:
an origin and the face normal. These pairs of origins/normals will be transformed
with all transformations on objects such that at a later stage, these
orinal references can be accessed. Then when a selection is requested, the boundaries
of objects will be computed together with their origin and normal. If an origin is close
to an original face surface and the normals are aligned, it is treated as part of the selection.

On top of that, EMerge has a selection interface that allows the user to select
faces and objects based on coordinates, layers etc.

In this demonstration we will create a complicated waveguide structure ending
in an air box to demonstrate the various selection methods.
"""

# First lets define some dimensions
mm = 0.001
wga = 22.86*mm
wgb = 10.16*mm
L = 50*mm

model = em.Simulation('BoundarySelectionDemo')
model.check_version("1.0.6") # Checks version compatibility.

# first lets define a WR90 waveguide
wg_box = em.geo.Box(L, wga, wgb, position=(-L, -wga/2, -wgb/2))
# Then define a capacitive iris cutout
cutout = em.geo.Box(2*mm, wga, wgb/2, position=(-L/2, -wga/2, -wgb/2))

# remove the cutout from the box. Notice that we use a different name.
# Geometry properties are not persistent after boolean operations so we 
# need the information of previous boxes.
wg_box_new = em.geo.remove(wg_box, cutout)

# define an air-box to radiat in.
airbox = em.geo.Box(L/2, L, L, position=(0,-L/2, -L/2))

# Now define the geometry
model.commit_geometry()

# Lets define a frequency range for our simulation. This is needed
# If we want to mesh our model.
model.mw.set_frequency_range(8e9, 10e9, 11)

# Now lets mesh our geometry
model.generate_mesh()

## We can now select faces and show them using the .view() interface

# The box is defined in XYZ space. The sides left/right correspond to the
# X-axis, The sides top/down to the Z-axis and front/back to the Y-axis.
# We have to provide which original object we want to pick the left side from.
feed_port = wg_box_new.face('left', tool=wg_box)

# We can also select the outside and exclude a given face. Because our airbox
# is not modified, we don't have to work with tools.
radiation_boundary = airbox.boundary(exclude='left')
# or one could use the no= optional argument in the face selection
radiation_boundary = airbox.face(no='left')

# Lets view our result
model.view(selections=[feed_port, radiation_boundary])

# As you can see, the appropriate faces have been selected.
# You can also see that the bottom side of the resultant box
# which has been split in two can still be selected because of the selection system.
model.view(selections=[wg_box_new.face('bottom', tool=wg_box),])

# You can also access faces of the original tool objects.
model.view(selections=[wg_box_new.face('left', tool=cutout),])

# Another way to select the radiation boundary on the right is by using the 
# selction interface.

# The interface works by a language like method-chaining philosophy.
# The .select attribute is a Selector class. The property .face returns
# The same selector class but with the 'face selection mode' turned on.
# Now we can call the 'inlayer' method which selects all faces of which the
# Center of mass is inside the layer ranging from the provided starting
# coordinate up to all coordinates that extend to the vector (L,0,0).
#
#            |                    |
#            |       ____\        |
#   (origin) + ---- vector ------>|
#            |                    |
#            |                    |
#            < inside is selected >
#

radiation_boundary_2 = model.select.face.inlayer(1*mm, 0,0, (L,0,0))
model.view(selections=[radiation_boundary_2,])

# Now lets define our simulation futher and do some farfield-computation!

port = model.mw.bc.ModalPort(feed_port, 1)
rad = model.mw.bc.AbsorbingBoundary(radiation_boundary)

model.mw.modal_analysis(port, 1)

# Run the simulation
data = model.mw.run_sweep()


# First the S11 plot
f = data.scalar.grid.freq
S11 = data.scalar.grid.S(1,1)

plot_sp(f, S11, labels=['S11'])

# And a far-field plot for demonstrative reasons.
# The θ=0 angle is defined as +x (1,0,0)
# the arc plane normal is the +z axis (0,0,1)

Efar = data.field[0].farfield_2d((1,0,0), (0,0,1), radiation_boundary)

# Finally we create the plot
plot_ff_polar(Efar.ang, Efar.normE)